# -*- coding: utf-8 -*-
"""
reV base gen and econ module.
"""
from abc import ABC, abstractmethod
import copy
from concurrent.futures import TimeoutError
import logging
import pandas as pd
import numpy as np
import os
import psutil
import json
import sys
from warnings import warn

from reV.config.project_points import ProjectPoints, PointsControl
from reV.handlers.outputs import Outputs
from reV.SAM.version_checker import PySamVersionChecker
from reV.utilities.exceptions import (OutputWarning, ExecutionError,
                                      ParallelExecutionWarning,
                                      OffshoreWindInputWarning)
from reV.utilities import log_versions, ModuleName

from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool

logger = logging.getLogger(__name__)


ATTR_DIR = os.path.dirname(os.path.realpath(__file__))
ATTR_DIR = os.path.join(ATTR_DIR, 'output_attributes')
with open(os.path.join(ATTR_DIR, 'other.json'), 'r') as f:
    OTHER_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, 'lcoe_fcr.json'), 'r') as f:
    LCOE_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, 'single_owner.json'), 'r') as f:
    SO_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, 'windbos.json'), 'r') as f:
    BOS_ATTRS = json.load(f)
with open(os.path.join(ATTR_DIR, 'lcoe_fcr_inputs.json'), 'r') as f:
    LCOE_IN_ATTRS = json.load(f)


class BaseGen(ABC):
    """Base class for reV gen and econ classes to run SAM simulations."""

    # Mapping of reV requests to SAM objects that should be used for simulation
    OPTIONS = {}

    # Mapping of reV generation / econ outputs to scale factors and units.
    OUT_ATTRS = copy.deepcopy(OTHER_ATTRS)

    # Mapping of reV econ outputs to scale factors and units.
    # Type is scalar or array and corresponds to the SAM single-site output
    # This is the OUT_ATTRS class attr for Econ but should also be accessible
    # to rev generation
    ECON_ATTRS = copy.deepcopy(OTHER_ATTRS)
    ECON_ATTRS.update(LCOE_ATTRS)
    ECON_ATTRS.update(SO_ATTRS)
    ECON_ATTRS.update(BOS_ATTRS)
    ECON_ATTRS.update(LCOE_IN_ATTRS)

    # SAM argument names used to calculate LCOE
    # Note that system_capacity is not included here because it is never used
    # downstream and could be confused with the supply_curve point capacity
    LCOE_ARGS = ('fixed_charge_rate', 'capital_cost', 'fixed_operating_cost',
                 'variable_operating_cost')

    def __init__(self, points_control, output_request, site_data=None,
                 out_fpath=None, drop_leap=False, mem_util_lim=0.4,
                 scale_outputs=True):
        """
        Parameters
        ----------
        points_control : reV.config.project_points.PointsControl
            Project points control instance for site and SAM config spec.
        output_request : list | tuple
            Output variables requested from SAM.
        site_data : str | pd.DataFrame | None
            Site-specific input data for SAM calculation. String should be a
            filepath that points to a csv, DataFrame is pre-extracted data.
            Rows match sites, columns are input keys. Need a "gid" column.
            Input as None if no site-specific data.
        out_fpath : str, optional
            Output .h5 file path, by default None
        drop_leap : bool
            Drop leap day instead of final day of year during leap years.
        mem_util_lim : float
            Memory utilization limit (fractional). This sets how many site
            results will be stored in-memory at any given time before flushing
            to disk.
        scale_outputs : bool
            Flag to scale outputs in-place immediately upon Gen returning data.
        """
        log_versions(logger)
        self._points_control = points_control
        self._year = None
        self._site_limit = None
        self._site_mem = None
        self._out_fpath = out_fpath
        self._meta = None
        self._time_index = None
        self._sam_module = None
        self._sam_obj_default = None
        self._drop_leap = drop_leap
        self.mem_util_lim = mem_util_lim
        self.scale_outputs = scale_outputs

        self._run_attrs = {'points_control': str(points_control),
                           'output_request': output_request,
                           'out_fpath': str(out_fpath),
                           'site_data': str(site_data),
                           'drop_leap': str(drop_leap),
                           'mem_util_lim': mem_util_lim,
                           }

        self._site_data = self._parse_site_data(site_data)
        self.add_site_data_to_pp(self._site_data)
        self._output_request = self._parse_output_request(output_request)

        # pre-initialize output arrays to store results when available.
        self._out = {}
        self._finished_sites = []
        self._out_n_sites = 0
        self._out_chunk = ()
        self._check_sam_version_inputs()

    @property
    def output_request(self):
        """Get the output variables requested from the user.

        Returns
        -------
        output_request : list
            Output variables requested from SAM.
        """
        return self._output_request

    @property
    def out_chunk(self):
        """Get the current output chunk index range (INCLUSIVE).

        Returns
        -------
        _out_chunk : tuple
            Two entry tuple (start, end) indicies (inclusive) for where the
            current data in-memory belongs in the final output.
        """
        return self._out_chunk

    @property
    def site_data(self):
        """Get the site-specific inputs in dataframe format.

        Returns
        -------
        _site_data : pd.DataFrame
            Site-specific input data for gen or econ calculation. Rows match
            sites, columns are variables.
        """
        return self._site_data

    @property
    def site_limit(self):
        """Get the number of sites results that can be stored in memory at once

        Returns
        -------
        _site_limit : int
            Number of site result sets that can be stored in memory at once
            without violating memory limits.
        """

        if self._site_limit is None:
            tot_mem = psutil.virtual_memory().total / 1e6
            avail_mem = self.mem_util_lim * tot_mem
            self._site_limit = int(np.floor(avail_mem / self.site_mem))
            logger.info('Limited to storing {0} sites in memory '
                        '({1:.1f} GB total hardware, {2:.1f} GB available '
                        'with {3:.1f}% utilization).'
                        .format(self._site_limit, tot_mem / 1e3,
                                avail_mem / 1e3, self.mem_util_lim * 100))

        return self._site_limit

    @property
    def site_mem(self):
        """Get the memory (MB) required to store all results for a single site.

        Returns
        -------
        _site_mem : float
            Memory (MB) required to store all results in requested in
            output_request for a single site.
        """

        if self._site_mem is None:
            # average the memory usage over n sites
            # (for better understanding of array overhead)
            n = 100
            self._site_mem = 0
            for request in self.output_request:
                dtype = 'float32'
                if request in self.OUT_ATTRS:
                    dtype = self.OUT_ATTRS[request].get('dtype', 'float32')

                shape = self._get_data_shape(request, n)
                self._site_mem += sys.getsizeof(np.ones(shape, dtype=dtype))

            self._site_mem = self._site_mem / 1e6 / n
            logger.info('Output results from a single site are calculated to '
                        'use {0:.1f} KB of memory.'
                        .format(self._site_mem / 1000))

        return self._site_mem

    @property
    def points_control(self):
        """Get project points controller.

        Returns
        -------
        points_control : reV.config.project_points.PointsControl
            Project points control instance for site and SAM config spec.
        """
        return self._points_control

    @property
    def project_points(self):
        """Get project points

        Returns
        -------
        project_points : reV.config.project_points.ProjectPoints
            Project points from the points control instance.
        """
        return self._points_control.project_points

    @property
    def sam_configs(self):
        """Get the sam config dictionary.

        Returns
        -------
        sam_configs : reV.config.sam.SAMGenConfig
            SAM config from the project points instance.
        """
        return self.project_points.sam_inputs

    @property
    def sam_metas(self):
        """
        SAM configurations including runtime module

        Returns
        -------
        sam_metas : dict
            Nested dictionary of SAM configuration files with module used
            at runtime
        """
        sam_metas = self.sam_configs.copy()
        for v in sam_metas.values():
            v.update({'module': self._sam_module.MODULE})

        return sam_metas

    @property
    def sam_module(self):
        """Get the SAM module class to be used for SAM simulations.

        Returns
        -------
        sam_module : object
            SAM object like PySAM.Pvwattsv7 or PySAM.Lcoefcr
        """
        return self._sam_module

    @property
    def out_fpath(self):
        """Get the output file path.

        Returns
        -------
        out_fpath : str | None
        """
        return self._out_fpath

    @property
    def meta(self):
        """Get resource meta for all sites in project points.

        Returns
        -------
        meta : pd.DataFrame
            Meta data df for sites in project points. Column names are meta
            data variables, rows are different sites. The row index
            does not indicate the site number if the project points are
            non-sequential or do not start from 0, so a 'gid' column is added.
        """
        return self._meta

    @property
    def time_index(self):
        """Get the resource time index data.

        Returns
        -------
        _time_index : pandas.DatetimeIndex
            Time-series datetime index
        """
        return self._time_index

    @property
    def run_attrs(self):
        """
        Run time attributes (__init__ args and kwargs)

        Returns
        -------
        run_attrs : dict
            Dictionary of runtime args and kwargs
        """
        return self._run_attrs

    @property
    def year(self):
        """Get the resource year.

        Returns
        -------
        _year : int
            Year of the time-series datetime index.
        """

        if self._year is None and self.time_index is not None:
            self._year = int(self.time_index.year[0])

        return self._year

    @property
    def tech(self):
        """Get the reV technology string.

        Returns
        -------
        tech : str
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam, econ)
            The string should be lower-cased with spaces and _ removed.
        """
        return self.project_points.tech

    @property
    def out(self):
        """Get the reV gen or econ output results.

        Returns
        -------
        out : dict
            Dictionary of gen or econ results from SAM.
        """
        out = {}
        for k, v in self._out.items():
            if k in self.OUT_ATTRS:
                scale_factor = self.OUT_ATTRS[k].get('scale_factor', 1)
            else:
                scale_factor = 1

            if scale_factor != 1 and self.scale_outputs:
                v = v.astype('float32')
                v /= scale_factor

            out[k] = v

        return out

    @out.setter
    def out(self, result):
        """Set the output attribute, unpack futures, clear output from mem.

        Parameters
        ----------
        result : list | dict | None
            Gen or Econ results to set to output dictionary. Use cases:
             - List input is interpreted as a futures list, which is unpacked
               before setting to the output dict.
             - Dictionary input is interpreted as an already unpacked result.
             - None is interpreted as a signal to clear the output dictionary.
        """
        if isinstance(result, list):
            # unpack futures list to dictionary first
            result = self.unpack_futures(result)

        if isinstance(result, dict):

            # iterate through dict where sites are keys and values are
            # corresponding results
            for site_gid, site_output in result.items():

                # check that the sites are stored sequentially then add to
                # the finished site list
                if self._finished_sites:
                    if int(site_gid) < np.max(self._finished_sites):
                        raise Exception('Site results are non sequential!')

                # unpack site output object
                self.unpack_output(site_gid, site_output)

                # add site gid to the finished list after outputs are unpacked
                self._finished_sites.append(site_gid)

        elif isinstance(result, type(None)):
            self._out.clear()
            self._finished_sites.clear()
        else:
            raise TypeError('Did not recognize the type of output. '
                            'Tried to set output type "{}", but requires '
                            'list, dict or None.'.format(type(result)))

    @staticmethod
    def _output_request_type_check(req):
        """Output request type check and ensure list for manipulation.

        Parameters
        ----------
        req : list | tuple | str
            Output request of variable type.

        Returns
        -------
        output_request : list
            Output request.
        """

        if isinstance(req, list):
            output_request = req
        elif isinstance(req, tuple):
            output_request = list(req)
        elif isinstance(req, str):
            output_request = [req]
        else:
            raise TypeError('Output request must be str, list, or tuple but '
                            'received: {}'.format(type(req)))

        return output_request

    @staticmethod
    def handle_leap_ti(ti, drop_leap=False):
        """Handle a time index for a leap year by dropping a day.

        Parameters
        ----------
        ti : pandas.DatetimeIndex
            Time-series datetime index with or without a leap day.
        drop_leap : bool
            Option to drop leap day (if True) or drop the last day of the year
            (if False).

        Returns
        -------
        ti : pandas.DatetimeIndex
            Time-series datetime index with length a multiple of 365.
        """

        # drop leap day or last day
        leap_day = ((ti.month == 2) & (ti.day == 29))
        last_day = ((ti.month == 12) & (ti.day == 31))
        if drop_leap:
            # preference is to drop leap day if exists
            ti = ti.drop(ti[leap_day])
        elif any(leap_day):
            # leap day exists but preference is to drop last day of year
            ti = ti.drop(ti[last_day])

        if len(ti) % 365 != 0:
            raise ValueError('Bad time index with length not a multiple of '
                             '365: {}'.format(ti))

        return ti

    @staticmethod
    def _pp_to_pc(points, points_range, sam_configs, tech,
                  sites_per_worker=None, res_file=None, curtailment=None):
        """
        Create ProjectControl from ProjectPoints

        Parameters
        ----------
        points : int | slice | list | str | pandas.DataFrame
                 | reV.config.project_points.PointsControl
            Single site integer,
            or slice or list specifying project points,
            or string pointing to a project points csv,
            or a pre-loaded project points DataFrame,
            or a fully instantiated PointsControl object.
        points_range : list | None
            Optional two-entry list specifying the index range of the sites to
            analyze. To be taken from the reV.config.PointsControl.split_range
            property.
        sam_configs : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s). Keys are the SAM
            config ID(s) which map to the config column in the project points
            CSV. Values are either a JSON SAM config file or dictionary of SAM
            config inputs. Can also be a single config file path or a
            pre loaded SAMConfig object.
        tech : str
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed.
        sites_per_worker : int
            Number of sites to run in series on a worker. None defaults to the
            resource file chunk size.
        res_file : str
            Filepath to single resource file, multi-h5 directory,
            or /h5_dir/prefix*suffix
        curtailment : NoneType | dict | str | config.curtailment.Curtailment
            Inputs for curtailment parameters. If not None, curtailment inputs
            are expected. Can be:
                - Explicit namespace of curtailment variables (dict)
                - Pointer to curtailment config json file with path (str)
                - Instance of curtailment config object
                  (config.curtailment.Curtailment)

        Returns
        -------
        pc : reV.config.project_points.PointsControl
            PointsControl object instance.
        """
        if not isinstance(points, ProjectPoints):
            # make Project Points instance
            pp = ProjectPoints(points, sam_configs, tech=tech,
                               res_file=res_file, curtailment=curtailment)
        else:
            pp = ProjectPoints(points.df, sam_configs, tech=tech,
                               res_file=res_file, curtailment=curtailment)

        #  make Points Control instance
        if points_range is not None:
            # PointsControl is for just a subset of the project points...
            # this is the case if generation is being initialized on one
            # of many HPC nodes in a large project
            pc = PointsControl.split(points_range[0], points_range[1], pp,
                                     sites_per_split=sites_per_worker)
        else:
            # PointsControl is for all of the project points
            pc = PointsControl(pp, sites_per_split=sites_per_worker)

        return pc

    @classmethod
    def get_pc(cls, points, points_range, sam_configs, tech,
               sites_per_worker=None, res_file=None, curtailment=None):
        """Get a PointsControl instance.

        Parameters
        ----------
        points : int | slice | list | str | pandas.DataFrame
                 | reV.config.project_points.PointsControl
            Single site integer,
            or slice or list specifying project points,
            or string pointing to a project points csv,
            or a pre-loaded project points DataFrame,
            or a fully instantiated PointsControl object.
        points_range : list | None
            Optional two-entry list specifying the index range of the sites to
            analyze. To be taken from the reV.config.PointsControl.split_range
            property.
        sam_configs : dict | str | SAMConfig
            SAM input configuration ID(s) and file path(s). Keys are the SAM
            config ID(s) which map to the config column in the project points
            CSV. Values are either a JSON SAM config file or dictionary of SAM
            config inputs. Can also be a single config file path or a
            pre loaded SAMConfig object.
        tech : str
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed.
        sites_per_worker : int
            Number of sites to run in series on a worker. None defaults to the
            resource file chunk size.
        res_file : str
            Filepath to single resource file, multi-h5 directory,
            or /h5_dir/prefix*suffix
        curtailment : NoneType | dict | str | config.curtailment.Curtailment
            Inputs for curtailment parameters. If not None, curtailment inputs
            are expected. Can be:
                - Explicit namespace of curtailment variables (dict)
                - Pointer to curtailment config json file with path (str)
                - Instance of curtailment config object
                  (config.curtailment.Curtailment)

        Returns
        -------
        pc : reV.config.project_points.PointsControl
            PointsControl object instance.
        """

        if tech not in cls.OPTIONS and tech.lower() != ModuleName.ECON:
            msg = ('Did not recognize reV-SAM technology string "{}". '
                   'Technology string options are: {}'
                   .format(tech, list(cls.OPTIONS.keys())))
            logger.error(msg)
            raise KeyError(msg)

        if sites_per_worker is None:
            # get the optimal sites per split based on res file chunk size
            sites_per_worker = cls.get_sites_per_worker(res_file)

        logger.debug('Sites per worker being set to {} for Gen/Econ '
                     'PointsControl.'.format(sites_per_worker))

        if isinstance(points, PointsControl):
            # received a pre-intialized instance of pointscontrol
            pc = points
        else:
            pc = cls._pp_to_pc(points, points_range, sam_configs, tech,
                               sites_per_worker=sites_per_worker,
                               res_file=res_file, curtailment=curtailment)

        return pc

    @staticmethod
    def get_sites_per_worker(res_file, default=100):
        """Get the nominal sites per worker (x-chunk size) for a given file.

        This is based on the concept that it is most efficient for one core to
        perform one read on one chunk of resource data, such that chunks will
        not have to be read into memory twice and no sites will be read
        redundantly.

        Parameters
        ----------
        res_file : str
            Filepath to single resource file, multi-h5 directory,
            or /h5_dir/prefix*suffix
        default : int
            Sites to be analyzed on a single core if the chunk size cannot be
            determined from res_file.

        Returns
        -------
        sites_per_worker : int
            Nominal sites to be analyzed per worker. This is set to the x-axis
            chunk size for windspeed and dni datasets for the WTK and NSRDB
            data, respectively.
        """
        if not res_file or not os.path.isfile(res_file):
            return default

        with Resource(res_file) as res:
            if 'wtk' in res_file.lower():
                for dset in res.datasets:
                    if 'speed' in dset:
                        # take nominal WTK chunks from windspeed
                        _, _, chunks = res.get_dset_properties(dset)
                        break
            elif 'nsrdb' in res_file.lower():
                # take nominal NSRDB chunks from dni
                _, _, chunks = res.get_dset_properties('dni')
            else:
                warn('Could not infer dataset chunk size as the resource type '
                     'could not be determined from the filename: {}'
                     .format(res_file))
                chunks = None

        if chunks is None:
            # if chunks not set, go to default
            sites_per_worker = default
            logger.debug('Sites per worker being set to {} (default) based on '
                         'no set chunk size in {}.'
                         .format(sites_per_worker, res_file))
        else:
            sites_per_worker = chunks[1]
            logger.debug('Sites per worker being set to {} based on chunk '
                         'size of {}.'.format(sites_per_worker, res_file))

        return sites_per_worker

    @staticmethod
    def unpack_futures(futures):
        """Combine list of futures results into their native dict format/type.

        Parameters
        ----------
        futures : list
            List of dictionary futures results.

        Returns
        -------
        out : dict
            Compiled results of the native future results type (dict).
        """

        out = {}
        for x in futures:
            out.update(x)

        return out

    @staticmethod
    @abstractmethod
    def run(points_control, tech=None, res_file=None, output_request=None,
            scale_outputs=True):
        """Run a reV-SAM analysis based on the points_control iterator.

        Parameters
        ----------
        points_control : reV.config.PointsControl
            A PointsControl instance dictating what sites and configs are run.
        tech : str
            SAM technology to analyze (pvwattsv7, windpower, tcsmoltensalt,
            solarwaterheat, troughphysicalheat, lineardirectsteam)
            The string should be lower-cased with spaces and _ removed.
        res_file : str
            Filepath to single resource file, multi-h5 directory,
            or /h5_dir/prefix*suffix
        output_request : list | tuple
            Output variables requested from SAM.
        scale_outputs : bool
            Flag to scale outputs in-place immediately upon returning data.

        Returns
        -------
        out : dict
            Output dictionary from the SAM reV_run function. Data is scaled
            within this function to the datatype specified in cls.OUT_ATTRS.
        """

    def _parse_site_data(self, inp):
        """Parse site-specific data from input arg

        Parameters
        ----------
        inp : str | pd.DataFrame | None
            Site data in .csv or pre-extracted dataframe format. None signifies
            that there is no extra site-specific data and that everything is
            fully defined in the input h5 and SAM json configs.

        Returns
        -------
        site_data : pd.DataFrame
            Site-specific data for econ calculation. Rows correspond to sites,
            columns are variables.
        """

        if inp is None or inp is False:
            # no input, just initialize dataframe with site gids as index
            site_data = pd.DataFrame(index=self.project_points.sites)
            site_data.index.name = 'gid'
        else:
            # explicit input, initialize df
            if isinstance(inp, str):
                if inp.endswith('.csv'):
                    site_data = pd.read_csv(inp)
            elif isinstance(inp, pd.DataFrame):
                site_data = inp
            else:
                # site data was not able to be set. Raise error.
                raise Exception('Site data input must be .csv or '
                                'dataframe, but received: {}'.format(inp))

            if 'gid' not in site_data and site_data.index.name != 'gid':
                # require gid as column label or index
                raise KeyError('Site data input must have "gid" column '
                               'to match reV site gid.')

            if site_data.index.name != 'gid':
                # make gid the dataframe index if not already
                site_data = site_data.set_index('gid', drop=True)

        if 'offshore' in site_data:
            if site_data['offshore'].sum() > 1:
                w = ('Found offshore sites in econ site data input. '
                     'This functionality has been deprecated. '
                     'Please run the reV offshore module to '
                     'calculate offshore wind lcoe.')
                warn(w, OffshoreWindInputWarning)
                logger.warning(w)

        return site_data

    def add_site_data_to_pp(self, site_data):
        """Add the site df (site-specific inputs) to project points dataframe.

        This ensures that only the relevant site's data will be passed through
        to parallel workers when points_control is iterated and split.

        Parameters
        ----------
        site_data : pd.DataFrame
            Site-specific data for econ calculation. Rows correspond to sites,
            columns are variables.
        """
        self.project_points.join_df(site_data, key=self.site_data.index.name)

    @abstractmethod
    def _parse_output_request(self, req):
        """Set the output variables requested from the user.

        Parameters
        ----------
        req : list | tuple
            Output variables requested from SAM.

        Returns
        -------
        output_request : list
            Output variables requested from SAM.
        """

    def _get_data_shape(self, dset, n_sites):
        """Get the output array shape based on OUT_ATTRS or PySAM.Outputs.

        Parameters
        ----------
        dset : str
            Variable name to get shape for.
        n_sites : int
            Number of sites for this data shape.

        Returns
        -------
        shape : tuple
            1D or 2D shape tuple for dset.
        """

        if dset in self.OUT_ATTRS:
            if self.OUT_ATTRS[dset]['type'] == 'array':
                data_shape = (len(self.time_index), n_sites)
            else:
                data_shape = (n_sites, )

        elif dset in self.project_points.all_sam_input_keys:
            data_shape = (n_sites, )
            data = list(self.project_points.sam_inputs.values())[0][dset]
            if isinstance(data, (list, tuple, np.ndarray, str)):
                msg = ('Cannot pass through non-scalar SAM input key "{}" '
                       'as an output_request!'.format(dset))
                logger.error(msg)
                raise ExecutionError(msg)

        else:
            if self._sam_obj_default is None:
                self._sam_obj_default = self.sam_module.default()

            try:
                out_data = getattr(self._sam_obj_default.Outputs, dset)
            except AttributeError as e:
                msg = ('Could not get data shape for dset "{}" '
                       'from object "{}". '
                       'Received the following error: "{}"'
                       .format(dset, self._sam_obj_default, e))
                logger.error(msg)
                raise ExecutionError(msg) from e
            else:
                if isinstance(out_data, (int, float, str)):
                    data_shape = (n_sites, )
                elif len(out_data) % len(self.time_index) == 0:
                    data_shape = (len(self.time_index), n_sites)
                else:
                    data_shape = (len(out_data), n_sites)

        return data_shape

    def _init_fpath(self):
        """Combine directory and filename, ensure .h5 ext., make out dirs."""

        if self._out_fpath is not None:

            # ensure output file is an h5
            if not self._out_fpath .endswith('.h5'):
                self._out_fpath += '.h5'

            # ensure year is in out_fpath
            if str(self.year) not in self._out_fpath:
                self._out_fpath = self._out_fpath.replace('.h5',
                                                          '_{}.h5'
                                                          .format(self.year))

            # create and use optional output dir
            dirout = os.path.dirname(self._out_fpath)
            if dirout and not os.path.exists(dirout):
                os.makedirs(dirout)

    def _init_h5(self, mode='w'):
        """Initialize the single h5 output file with all output requests.

        Parameters
        ----------
        mode : str
            Mode to instantiate h5py.File instance
        """

        if self._out_fpath is not None:

            if 'w' in mode:
                logger.info('Initializing full output file: "{}" with mode: {}'
                            .format(self._out_fpath, mode))
            elif 'a' in mode:
                logger.info('Appending data to output file: "{}" with mode: {}'
                            .format(self._out_fpath, mode))

            attrs = {d: {} for d in self.output_request}
            chunks = {}
            dtypes = {}
            shapes = {}

            # flag to write time index if profiles are being output
            write_ti = False

            for dset in self.output_request:

                tmp = 'other'
                if dset in self.OUT_ATTRS:
                    tmp = dset

                attrs[dset]['units'] = self.OUT_ATTRS[tmp].get('units',
                                                               'unknown')
                attrs[dset]['scale_factor'] = \
                    self.OUT_ATTRS[tmp].get('scale_factor', 1)
                chunks[dset] = self.OUT_ATTRS[tmp].get('chunks', None)
                dtypes[dset] = self.OUT_ATTRS[tmp].get('dtype', 'float32')
                shapes[dset] = self._get_data_shape(dset, len(self.meta))
                if len(shapes[dset]) > 1:
                    write_ti = True

            # only write time index if profiles were found in output request
            if write_ti:
                ti = self.time_index
            else:
                ti = None

            Outputs.init_h5(self._out_fpath, self.output_request, shapes,
                            attrs, chunks, dtypes, self.meta, time_index=ti,
                            configs=self.sam_metas, run_attrs=self.run_attrs,
                            mode=mode)

    def _init_out_arrays(self, index_0=0):
        """Initialize output arrays based on the number of sites that can be
        stored in memory safely.

        Parameters
        ----------
        index_0 : int
            This is the site list index (not gid) for the first site in the
            output data. If a node cannot process all sites in-memory at once,
            this is used to segment the sites in the current output chunk.
        """

        self._out = {}
        self._finished_sites = []

        # Output chunk is the index range (inclusive) of this set of site outs
        self._out_chunk = (index_0, np.min((index_0 + self.site_limit,
                                            len(self.project_points) - 1)))
        self._out_n_sites = int(self.out_chunk[1] - self.out_chunk[0]) + 1

        logger.info('Initializing in-memory outputs for {} sites with gids '
                    '{} through {} inclusive (site list index {} through {})'
                    .format(self._out_n_sites,
                            self.project_points.sites[self.out_chunk[0]],
                            self.project_points.sites[self.out_chunk[1]],
                            self.out_chunk[0], self.out_chunk[1]))

        for request in self.output_request:
            dtype = 'float32'
            if request in self.OUT_ATTRS and self.scale_outputs:
                dtype = self.OUT_ATTRS[request].get('dtype', 'float32')

            shape = self._get_data_shape(request, self._out_n_sites)
            # initialize the output request as an array of zeros
            self._out[request] = np.zeros(shape, dtype=dtype)

    def _check_sam_version_inputs(self):
        """Check the PySAM version and input keys. Fix where necessary."""
        for key, parameters in self.project_points.sam_inputs.items():
            updated = PySamVersionChecker.run(self.tech, parameters)
            sam_obj = self._points_control._project_points._sam_config_obj
            sam_obj._inputs[key] = updated

    def unpack_output(self, site_gid, site_output):
        """Unpack a SAM SiteOutput object to the output attribute.

        Parameters
        ----------
        site_gid : int
            Resource-native site gid (index).
        site_output : dict
            SAM site output object.
        """

        # iterate through the site results
        for var, value in site_output.items():
            if var not in self._out:
                raise KeyError('Tried to collect output variable "{}", but it '
                               'was not yet initialized in the output '
                               'dictionary.')

            # get the index in the output array for the current site
            i = self.site_index(site_gid, out_index=True)

            # check to see if we have exceeded the current output chunk.
            # If so, flush data to disk and reset the output initialization
            if i + 1 > self._out_n_sites:
                self.flush()
                global_site_index = self.site_index(site_gid)
                self._init_out_arrays(index_0=global_site_index)
                i = self.site_index(site_gid, out_index=True)

            if isinstance(value, (list, tuple, np.ndarray)):
                if not isinstance(value, np.ndarray):
                    value = np.array(value)

                self._out[var][:, i] = value.T

            elif value != 0:
                self._out[var][i] = value

    def site_index(self, site_gid, out_index=False):
        """Get the index corresponding to the site gid.

        Parameters
        ----------
        site_gid : int
            Resource-native site index (gid).
        out_index : bool
            Option to get output index (if true) which is the column index in
            the current in-memory output array, or (if false) the global site
            index from the project points site list.

        Returns
        -------
        index : int
            Global site index if out_index=False, otherwise column index in
            the current in-memory output array.
        """

        # get the index for site_gid in the (global) project points site list.
        global_site_index = self.project_points.sites.index(site_gid)

        if not out_index:
            output_index = global_site_index
        else:
            output_index = global_site_index - self.out_chunk[0]
            if output_index < 0:
                raise ValueError('Attempting to set output data for site with '
                                 'gid {} to global site index {}, which was '
                                 'already set based on the current output '
                                 'index chunk of {}'
                                 .format(site_gid, global_site_index,
                                         self.out_chunk))

        return output_index

    def flush(self):
        """Flush the output data in self.out attribute to disk in .h5 format.

        The data to be flushed is accessed from the instance attribute
        "self.out". The disk target is based on the instance attributes
        "self._out_fpath". Data is not flushed if _fpath is None or if .out is
        empty.
        """

        # handle output file request if file is specified and .out is not empty
        if isinstance(self._out_fpath, str) and self._out:
            logger.info('Flushing outputs to disk, target file: "{}"'
                        .format(self._out_fpath))

            # get the slice of indices to write outputs to
            islice = slice(self.out_chunk[0], self.out_chunk[1] + 1)

            # open output file in append mode to add output results to
            with Outputs(self._out_fpath, mode='a') as f:

                # iterate through all output requests writing each as a dataset
                for dset, arr in self._out.items():
                    if len(arr.shape) == 1:
                        # write array of scalars
                        f[dset, islice] = arr
                    else:
                        # write 2D array of profiles
                        f[dset, :, islice] = arr

            logger.debug('Flushed output successfully to disk.')

    def _pre_split_pc(self, pool_size=(os.cpu_count() * 2)):
        """Pre-split project control iterator into sub chunks to further
        split the parallelization.

        Parameters
        ----------
        pool_size : int
            Number of futures to submit to a single process pool for
            parallel futures.

        Returns
        -------
        N : int
            Total number of points control split instances.
        pc_chunks : list
            List of lists of points control split instances.
        """
        N = 0
        pc_chunks = []
        i_chunk = []

        for i, split in enumerate(self.points_control):
            N += 1
            i_chunk.append(split)
            if (i + 1) % pool_size == 0:
                pc_chunks.append(i_chunk)
                i_chunk = []

        if i_chunk:
            pc_chunks.append(i_chunk)

        logger.debug('Pre-splitting points control into {} chunks with the '
                     'following chunk sizes: {}'
                     .format(len(pc_chunks), [len(x) for x in pc_chunks]))
        return N, pc_chunks

    def _parallel_run(self, max_workers=None, pool_size=(os.cpu_count() * 2),
                      timeout=1800, **kwargs):
        """Execute parallel compute.

        Parameters
        ----------
        max_workers : None | int
            Number of workers. None will default to cpu count.
        pool_size : int
            Number of futures to submit to a single process pool for
            parallel futures.
        timeout : int | float
            Number of seconds to wait for parallel run iteration to complete
            before returning zeros.
        kwargs : dict
            Keyword arguments to self.run().
        """

        max_workers = os.cpu_count() if max_workers is None else max_workers
        logger.info('Running parallel execution with max_workers={}'
                    .format(max_workers))
        i = 0
        N, pc_chunks = self._pre_split_pc(pool_size=pool_size)
        for j, pc_chunk in enumerate(pc_chunks):
            logger.debug('Starting process pool for points control '
                         'iteration {} out of {}'
                         .format(j + 1, len(pc_chunks)))

            failed_futures = False
            chunks = {}
            futures = []
            loggers = [__name__, 'reV.gen', 'reV.econ', 'reV']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                for pc in pc_chunk:
                    future = exe.submit(self.run, pc, **kwargs)
                    futures.append(future)
                    chunks[future] = pc

                for future in futures:
                    i += 1
                    try:
                        result = future.result(timeout=timeout)
                    except TimeoutError:
                        failed_futures = True
                        sites = chunks[future].project_points.sites
                        result = self._handle_failed_future(future, i, sites,
                                                            timeout)

                    self.out = result

                    mem = psutil.virtual_memory()
                    m = ('Parallel run at iteration {0} out of {1}. '
                         'Memory utilization is {2:.3f} GB out of {3:.3f} GB '
                         'total ({4:.1f}% used, intended limit of {5:.1f}%)'
                         .format(i, N, mem.used / 1e9, mem.total / 1e9,
                                 100 * mem.used / mem.total,
                                 100 * self.mem_util_lim))
                    logger.info(m)

                if failed_futures:
                    logger.info('Forcing pool shutdown after failed futures.')
                    exe.shutdown(wait=False)
                    logger.info('Forced pool shutdown complete.')

        self.flush()

    def _handle_failed_future(self, future, i, sites, timeout):
        """Handle a failed future and return zeros.

        Parameters
        ----------
        future : concurrent.futures.Future
            Failed future to cancel.
        i : int
            Iteration number for logging
        sites : list
            List of site gids belonging to this failed future.
        timeout : int
            Number of seconds to wait for parallel run iteration to complete
            before returning zeros.
        """

        w = ('Iteration {} hit the timeout limit of {} seconds! Passing zeros.'
             .format(i, timeout))
        logger.warning(w)
        warn(w, OutputWarning)

        site_out = {k: 0 for k in self.output_request}
        result = {site: site_out for site in sites}

        try:
            cancelled = future.cancel()
        except Exception as e:
            w = 'Could not cancel future! Received exception: {}'.format(e)
            logger.warning(w)
            warn(w, ParallelExecutionWarning)

        if not cancelled:
            w = 'Could not cancel future!'
            logger.warning(w)
            warn(w, ParallelExecutionWarning)

        return result



# -*- coding: utf-8 -*-
"""reV-to-SAM generation interface module.

Wraps the NREL-PySAM pvwattsv5, windpower, and tcsmolensalt modules with
additional reV features.
"""
from abc import ABC, abstractmethod
import copy
import os
import logging
import numpy as np
import pandas as pd
from warnings import warn
import PySAM.Pvwattsv5 as PySamPv5
import PySAM.Pvwattsv7 as PySamPv7
import PySAM.Pvsamv1 as PySamDetailedPv
import PySAM.Windpower as PySamWindPower
import PySAM.TcsmoltenSalt as PySamCSP
import PySAM.Swh as PySamSwh
import PySAM.TroughPhysicalProcessHeat as PySamTpph
import PySAM.LinearFresnelDsgIph as PySamLds
import PySAM.MhkWave as PySamMhkWave

from reV.SAM.defaults import (DefaultPvWattsv5,
                              DefaultPvWattsv7,
                              DefaultPvSamv1,
                              DefaultWindPower,
                              DefaultTcsMoltenSalt,
                              DefaultSwh,
                              DefaultTroughPhysicalProcessHeat,
                              DefaultLinearFresnelDsgIph,
                              DefaultMhkWave)
from reV.utilities.exceptions import (SAMInputWarning, SAMExecutionError,
                                      InputError)
from reV.utilities.curtailment import curtail
from reV.SAM.SAM import RevPySam
from reV.SAM.econ import LCOE, SingleOwner

logger = logging.getLogger(__name__)


class AbstractSamGeneration(RevPySam, ABC):
    """Base class for SAM generation simulations."""

    def __init__(self, resource, meta, sam_sys_inputs, site_sys_inputs=None,
                 output_request=None, drop_leap=False):
        """Initialize a SAM generation object.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        output_request : list
            Requested SAM outputs (e.g., 'cf_mean', 'annual_energy',
            'cf_profile', 'gen_profile', 'energy_yield', 'ppa_price',
            'lcoe_fcr').
        drop_leap : bool
            Drops February 29th from the resource data. If False, December
            31st is dropped from leap years.
        """

        # drop the leap day
        if drop_leap:
            resource = self.drop_leap(resource)

        # make sure timezone and elevation are in the meta data
        meta = self.tz_elev_check(sam_sys_inputs, site_sys_inputs, meta)

        # don't pass resource to base class,
        # set in concrete generation classes instead
        super().__init__(meta, sam_sys_inputs, output_request,
                         site_sys_inputs=site_sys_inputs)

        # Set the site number using resource
        if hasattr(resource, 'name'):
            self._site = resource.name
        else:
            self._site = None

        self.check_resource_data(resource)
        self.set_resource_data(resource, meta)

    @classmethod
    def _get_res(cls, res_df, output_request):
        """Get the resource arrays and pass through for output (single site).

        Parameters
        ----------
        res_df : pd.DataFrame
            2D table with resource data.
        output_request : list
            Outputs to retrieve from SAM.

        Returns
        -------
        res_mean : dict | None
            Dictionary object with variables for resource arrays.
        out_req_cleaned : list
            Output request list with the resource request entries removed.
        """

        out_req_cleaned = copy.deepcopy(output_request)
        res_out = None

        res_reqs = []
        ti = res_df.index
        for req in out_req_cleaned:
            if req in res_df:
                res_reqs.append(req)
                if res_out is None:
                    res_out = {}
                res_out[req] = cls.ensure_res_len(res_df[req].values, ti)

        for req in res_reqs:
            out_req_cleaned.remove(req)

        return res_out, out_req_cleaned

    @staticmethod
    def _get_res_mean(resource, res_gid, output_request):
        """Get the resource annual means (single site).

        Parameters
        ----------
        resource : rex.sam_resource.SAMResource
            SAM resource object for WIND resource
        res_gid : int
            Site to extract means for
        output_request : list
            Outputs to retrieve from SAM.

        Returns
        -------
        res_mean : dict | None
            Dictionary object with variables for resource means.
        out_req_nomeans : list
            Output request list with the resource mean entries removed.
        """

        out_req_nomeans = copy.deepcopy(output_request)
        res_mean = None
        idx = resource.sites.index(res_gid)
        irrad_means = ('dni_mean', 'dhi_mean', 'ghi_mean')

        if 'ws_mean' in out_req_nomeans:
            out_req_nomeans.remove('ws_mean')
            res_mean = {}
            res_mean['ws_mean'] = resource['mean_windspeed', idx]

        else:
            for var in resource.var_list:
                label_1 = '{}_mean'.format(var)
                label_2 = 'mean_{}'.format(var)
                if label_1 in out_req_nomeans:
                    out_req_nomeans.remove(label_1)
                    if res_mean is None:
                        res_mean = {}
                    res_mean[label_1] = resource[label_2, idx]

                    if label_1 in irrad_means:
                        # convert to kWh/m2/day
                        res_mean[label_1] /= 1000
                        res_mean[label_1] *= 24

        return res_mean, out_req_nomeans

    def check_resource_data(self, resource):
        """Check resource dataframe for NaN values

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        """
        if pd.isna(resource).any().any():
            bad_vars = pd.isna(resource).any(axis=0)
            bad_vars = resource.columns[bad_vars].values.tolist()
            msg = ('Found NaN values for site {} in variables {}'
                   .format(self.site, bad_vars))
            logger.error(msg)
            raise InputError(msg)

    @abstractmethod
    def set_resource_data(self, resource, meta):
        """Placeholder for resource data setting (nsrdb or wtk)"""

    @staticmethod
    def tz_elev_check(sam_sys_inputs, site_sys_inputs, meta):
        """Check timezone+elevation input and use json config
        timezone+elevation if not in resource meta.

        Parameters
        ----------
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        meta : pd.DataFrame
            1D table with resource meta data.

        Returns
        -------
        meta : pd.DataFrame
            1D table with resource meta data. Will include "timezone"
            and "elevation" from the sam and site system inputs if found.
        """

        if meta is not None:
            if sam_sys_inputs is not None:
                if 'elevation' in sam_sys_inputs:
                    meta['elevation'] = sam_sys_inputs['elevation']
                if 'timezone' in sam_sys_inputs:
                    meta['timezone'] = int(sam_sys_inputs['timezone'])

            # site-specific inputs take priority over generic system inputs
            if site_sys_inputs is not None:
                if 'elevation' in site_sys_inputs:
                    meta['elevation'] = site_sys_inputs['elevation']
                if 'timezone' in site_sys_inputs:
                    meta['timezone'] = int(site_sys_inputs['timezone'])

            if 'timezone' not in meta:
                msg = ('Need timezone input to run SAM gen. Not found in '
                       'resource meta or technology json input config.')
                raise SAMExecutionError(msg)

        return meta

    @property
    def has_timezone(self):
        """ Returns true if instance has a timezone set """
        if self._meta is not None:
            if 'timezone' in self.meta:
                return True

        return False

    def cf_mean(self):
        """Get mean capacity factor (fractional) from SAM.

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
        """
        return self['capacity_factor'] / 100

    def cf_profile(self):
        """Get hourly capacity factor (frac) profile in orig timezone.

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return self.gen_profile() / self.sam_sys_inputs['system_capacity']

    def annual_energy(self):
        """Get annual energy generation value in kWh from SAM.

        Returns
        -------
        output : float
            Annual energy generation (kWh).
        """
        return self['annual_energy']

    def energy_yield(self):
        """Get annual energy yield value in kwh/kw from SAM.

        Returns
        -------
        output : float
            Annual energy yield (kwh/kw).
        """
        return self['kwh_per_kw']

    def gen_profile(self):
        """Get power generation profile (orig timezone) in kW.

        Returns
        -------
        output : np.ndarray
            1D array of hourly power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self['gen'], dtype=np.float32)

    def collect_outputs(self, output_lookup=None):
        """
        Collect SAM gen output_request. Rolls outputs to UTC if appropriate.

        Parameters
        ----------
        output_lookup : dict | None
            Lookup dictionary mapping output keys to special output methods.
            None defaults to generation default outputs.
        """

        if output_lookup is None:
            output_lookup = {'cf_mean': self.cf_mean,
                             'cf_profile': self.cf_profile,
                             'annual_energy': self.annual_energy,
                             'energy_yield': self.energy_yield,
                             'gen_profile': self.gen_profile,
                             }

        super().collect_outputs(output_lookup=output_lookup)

    def run_gen_and_econ(self):
        """Run SAM generation with possibility for follow on econ analysis."""

        lcoe_out_reqs = None
        so_out_reqs = None
        lcoe_vars = ('lcoe_fcr', 'fixed_charge_rate', 'capital_cost',
                     'fixed_operating_cost', 'variable_operating_cost')
        so_vars = ('ppa_price', 'lcoe_real', 'lcoe_nom',
                   'project_return_aftertax_npv', 'flip_actual_irr',
                   'gross_revenue')
        if 'lcoe_fcr' in self.output_request:
            lcoe_out_reqs = [r for r in self.output_request if r in lcoe_vars]
            self.output_request = [r for r in self.output_request
                                   if r not in lcoe_out_reqs]
        elif any(x in self.output_request for x in so_vars):
            so_out_reqs = [r for r in self.output_request if r in so_vars]
            self.output_request = [r for r in self.output_request
                                   if r not in so_out_reqs]

        # Execute the SAM generation compute module (pvwattsv7, windpower, etc)
        self.run()

        # Execute a follow-on SAM econ compute module
        # (lcoe_fcr, singleowner, etc)
        if lcoe_out_reqs is not None:
            self.sam_sys_inputs['annual_energy'] = self.annual_energy()
            lcoe = LCOE(self.sam_sys_inputs, output_request=lcoe_out_reqs)
            lcoe.assign_inputs()
            lcoe.execute()
            lcoe.collect_outputs()
            lcoe.outputs_to_utc_arr()
            self.outputs.update(lcoe.outputs)

        elif so_out_reqs is not None:
            self.sam_sys_inputs['gen'] = self.gen_profile()
            so = SingleOwner(self.sam_sys_inputs, output_request=so_out_reqs)
            so.assign_inputs()
            so.execute()
            so.collect_outputs()
            so.outputs_to_utc_arr()
            self.outputs.update(so.outputs)

    def run(self):
        """Run a reV-SAM generation object by assigning inputs, executing the
        SAM simulation, collecting outputs, and converting all arrays to UTC.
        """
        self.assign_inputs()
        self.execute()
        self.collect_outputs()
        self.outputs_to_utc_arr()

    @classmethod
    def reV_run(cls, points_control, res_file, site_df,
                output_request=('cf_mean',), drop_leap=False,
                gid_map=None):
        """Execute SAM generation based on a reV points control instance.

        Parameters
        ----------
        points_control : config.PointsControl
            PointsControl instance containing project points site and SAM
            config info.
        res_file : str
            Resource file with full path.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        output_request : list | tuple
            Outputs to retrieve from SAM.
        drop_leap : bool
            Drops February 29th from the resource data. If False, December
            31st is dropped from leap years.
        gid_map : None | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This enables the user to input unique
            generation gids in the project points that map to non-unique
            resource gids. This can be None or a pre-extracted dict.

        Returns
        -------
        out : dict
            Nested dictionaries where the top level key is the site index,
            the second level key is the variable name, second level value is
            the output variable value.
        """
        # initialize output dictionary
        out = {}

        # Get the RevPySam resource object
        resources = RevPySam.get_sam_res(res_file,
                                         points_control.project_points,
                                         points_control.project_points.tech,
                                         output_request=output_request,
                                         gid_map=gid_map)

        # run resource through curtailment filter if applicable
        curtailment = points_control.project_points.curtailment
        if curtailment is not None:
            resources = curtail(resources, curtailment,
                                random_seed=curtailment.random_seed)

        # iterate through project_points gen_gid values
        for gen_gid in points_control.project_points.sites:

            # Lookup the resource gid if there's a mapping and get the resource
            # data from the SAMResource object using the res_gid.
            res_gid = gen_gid if gid_map is None else gid_map[gen_gid]
            site_res_df, site_meta = resources._get_res_df(res_gid)

            # drop the leap day
            if drop_leap:
                site_res_df = cls.drop_leap(site_res_df)

            _, inputs = points_control.project_points[gen_gid]

            # get resource data pass-throughs and resource means
            res_outs, out_req_cleaned = cls._get_res(site_res_df,
                                                     output_request)
            res_mean, out_req_cleaned = cls._get_res_mean(resources, res_gid,
                                                          out_req_cleaned)

            # iterate through requested sites.
            sim = cls(resource=site_res_df, meta=site_meta,
                      sam_sys_inputs=inputs, output_request=out_req_cleaned,
                      site_sys_inputs=dict(site_df.loc[gen_gid, :]))
            sim.run_gen_and_econ()

            # collect outputs to dictout
            out[gen_gid] = sim.outputs

            if res_outs is not None:
                out[gen_gid].update(res_outs)

            if res_mean is not None:
                out[gen_gid].update(res_mean)

        return out


class AbstractSamSolar(AbstractSamGeneration, ABC):
    """Base Class for Solar generation from SAM"""

    def set_resource_data(self, resource, meta):
        """Set NSRDB resource data arrays.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        """

        time_index = resource.index
        self.time_interval = self.get_time_interval(resource.index.values)

        # map resource data names to SAM required data names
        var_map = {'dni': 'dn',
                   'dhi': 'df',
                   'ghi': 'gh',
                   'clearskydni': 'dn',
                   'clearskydhi': 'df',
                   'clearskyghi': 'gh',
                   'windspeed': 'wspd',
                   'airtemperature': 'tdry',
                   'temperature': 'tdry',
                   'temp': 'tdry',
                   'dewpoint': 'tdew',
                   'surfacepressure': 'pres',
                   'pressure': 'pres',
                   'surfacealbedo': 'albedo',
                   }
        lower_case = {k: k.lower().replace(' ', '').replace('_', '')
                      for k in resource.columns}
        irrad_vars = ['dn', 'df', 'gh']

        resource = resource.rename(mapper=lower_case, axis='columns')
        resource = resource.rename(mapper=var_map, axis='columns')
        time_index = resource.index
        resource = {k: np.array(v) for (k, v) in
                    resource.to_dict(orient='list').items()}

        # set resource variables
        for var, arr in resource.items():
            if var != 'time_index':

                # ensure that resource array length is multiple of 8760
                arr = np.roll(
                    self.ensure_res_len(arr, time_index),
                    int(self._meta['timezone'] * self.time_interval))

                if var in irrad_vars:
                    if np.min(arr) < 0:
                        warn('Solar irradiance variable "{}" has a minimum '
                             'value of {}. Truncating to zero.'
                             .format(var, np.min(arr)), SAMInputWarning)
                        arr = np.where(arr < 0, 0, arr)

                resource[var] = arr.tolist()

        resource['lat'] = meta['latitude']
        resource['lon'] = meta['longitude']
        resource['tz'] = meta['timezone']

        if 'elevation' in meta:
            resource['elev'] = meta['elevation']
        else:
            resource['elev'] = 0.0

        time_index = self.ensure_res_len(time_index, time_index)
        resource['minute'] = time_index.minute
        resource['hour'] = time_index.hour
        resource['year'] = time_index.year
        resource['month'] = time_index.month
        resource['day'] = time_index.day

        if 'albedo' in resource:
            self['albedo'] = resource.pop('albedo')

        self['solar_resource_data'] = resource


class AbstractSamPv(AbstractSamSolar, ABC):
    """Photovoltaic (PV) generation with either pvwatts of detailed pv.
    """

    # set these class attrs in concrete subclasses
    MODULE = None
    PYSAM = None

    def __init__(self, resource, meta, sam_sys_inputs, site_sys_inputs=None,
                 output_request=None, drop_leap=False):
        """Initialize a SAM solar object.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        output_request : list
            Requested SAM outputs (e.g., 'cf_mean', 'annual_energy',
            'cf_profile', 'gen_profile', 'energy_yield', 'ppa_price',
            'lcoe_fcr').
        drop_leap : bool
            Drops February 29th from the resource data. If False, December
            31st is dropped from leap years.
        """

        # need to check tilt=lat and azimuth for pv systems
        sam_sys_inputs = self.set_latitude_tilt_az(sam_sys_inputs, meta)

        super().__init__(resource, meta, sam_sys_inputs,
                         site_sys_inputs=site_sys_inputs,
                         output_request=output_request,
                         drop_leap=drop_leap)

    @staticmethod
    def set_latitude_tilt_az(sam_sys_inputs, meta):
        """Check if tilt is specified as latitude and set tilt=lat, az=180 or 0

        Parameters
        ----------
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        meta : pd.DataFrame
            1D table with resource meta data.

        Returns
        -------
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
            If for a pv simulation the "tilt" parameter was originally not
            present or set to 'lat' or 'latitude', the tilt will be set to
            the absolute value of the latitude found in meta and the azimuth
            will be 180 if lat>0, 0 if lat<0.
        """

        set_tilt = False
        if sam_sys_inputs is not None and meta is not None:
            if 'tilt' not in sam_sys_inputs:
                warn('No tilt specified, setting at latitude.',
                     SAMInputWarning)
                set_tilt = True
            else:
                if (sam_sys_inputs['tilt'] == 'lat'
                        or sam_sys_inputs['tilt'] == 'latitude'):
                    set_tilt = True

        if set_tilt:
            # set tilt to abs(latitude)
            sam_sys_inputs['tilt'] = np.abs(meta['latitude'])
            if meta['latitude'] > 0:
                # above the equator, az = 180
                sam_sys_inputs['azimuth'] = 180
            else:
                # below the equator, az = 0
                sam_sys_inputs['azimuth'] = 0

            logger.debug('Tilt specified at "latitude", setting tilt to: {}, '
                         'azimuth to: {}'
                         .format(sam_sys_inputs['tilt'],
                                 sam_sys_inputs['azimuth']))
        return sam_sys_inputs

    def cf_mean(self):
        """Get mean capacity factor (fractional) from SAM.

        NOTE: PV capacity factor is the AC power production / the DC nameplate

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
            PV CF is calculated as AC power / DC nameplate.
        """
        return self['capacity_factor'] / 100

    def cf_profile(self):
        """Get hourly capacity factor (frac) profile in orig timezone.

        NOTE: PV capacity factor is the AC power production / the DC nameplate

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
            PV CF is calculated as AC power / DC nameplate.
        """
        return self.gen_profile() / self.sam_sys_inputs['system_capacity']

    def gen_profile(self):
        """Get AC inverter power generation profile (orig timezone) in kW.
        This is an alias of the "ac" SAM output variable.

        Returns
        -------
        output : np.ndarray
            1D array of AC inverter power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return self.ac()

    def ac(self):
        """Get AC inverter power generation profile (orig timezone) in kW.

        Returns
        -------
        output : np.ndarray
            1D array of AC inverter power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self['ac'], dtype=np.float32) / 1000

    def dc(self):
        """
        Get DC array power generation profile (orig timezone) in kW.

        Returns
        -------
        output : np.ndarray
            1D array of DC array power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self['dc'], dtype=np.float32) / 1000

    def clipped_power(self):
        """
        Get the clipped DC power generated behind the inverter
        (orig timezone) in kW.

        Returns
        -------
        clipped : np.ndarray
            1D array of clipped DC power in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        ac = self.ac()
        dc = self.dc()

        return np.where(ac < ac.max(), 0, dc - ac)

    @staticmethod
    @abstractmethod
    def default():
        """Get the executed default pysam object."""

    def collect_outputs(self, output_lookup=None):
        """Collect SAM gen output_request.

        Parameters
        ----------
        output_lookup : dict | None
            Lookup dictionary mapping output keys to special output methods.
            None defaults to generation default outputs.
        """

        if output_lookup is None:
            output_lookup = {'cf_mean': self.cf_mean,
                             'cf_profile': self.cf_profile,
                             'annual_energy': self.annual_energy,
                             'energy_yield': self.energy_yield,
                             'gen_profile': self.gen_profile,
                             'ac': self.ac,
                             'dc': self.dc,
                             'clipped_power': self.clipped_power
                             }

        super().collect_outputs(output_lookup=output_lookup)


class PvWattsv5(AbstractSamPv):
    """Photovoltaic (PV) generation with pvwattsv5.
    """
    MODULE = 'pvwattsv5'
    PYSAM = PySamPv5

    @staticmethod
    def default():
        """Get the executed default pysam PVWATTSV5 object.

        Returns
        -------
        PySAM.Pvwattsv5
        """
        return DefaultPvWattsv5.default()


class PvWattsv7(AbstractSamPv):
    """Photovoltaic (PV) generation with pvwattsv7.
    """
    MODULE = 'pvwattsv7'
    PYSAM = PySamPv7

    @staticmethod
    def default():
        """Get the executed default pysam PVWATTSV7 object.

        Returns
        -------
        PySAM.Pvwattsv7
        """
        return DefaultPvWattsv7.default()


class PvSamv1(AbstractSamPv):
    """Detailed PV model"""

    MODULE = 'Pvsamv1'
    PYSAM = PySamDetailedPv

    def ac(self):
        """Get AC inverter power generation profile (orig timezone) in kW.

        Returns
        -------
        output : np.ndarray
            1D array of AC inverter power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self['gen'], dtype=np.float32)

    def dc(self):
        """
        Get DC array power generation profile (orig timezone) in kW.

        Returns
        -------
        output : np.ndarray
            1D array of DC array power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self['dc_net'], dtype=np.float32)

    @staticmethod
    def default():
        """Get the executed default pysam Pvsamv1 object.

        Returns
        -------
        PySAM.Pvsamv1
        """
        return DefaultPvSamv1.default()


class TcsMoltenSalt(AbstractSamSolar):
    """Concentrated Solar Power (CSP) generation with tower molten salt
    """
    MODULE = 'tcsmolten_salt'
    PYSAM = PySamCSP

    def cf_profile(self):
        """Get absolute value hourly capacity factor (frac) profile in
        orig timezone.

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
        """
        x = np.abs(self.gen_profile() / self.sam_sys_inputs['system_capacity'])
        return x

    @staticmethod
    def default():
        """Get the executed default pysam CSP object.

        Returns
        -------
        PySAM.TcsmoltenSalt
        """
        return DefaultTcsMoltenSalt.default()


class AbstractSamSolarThermal(AbstractSamSolar, ABC):
    """Base class for solar thermal """
    PYSAM_WEATHER_TAG = None

    def set_resource_data(self, resource, meta):
        """
        Set NSRDB resource file. Overloads Solar.set_resource_data(). Solar
        thermal PySAM models require a data file, not raw data.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        """
        self.time_interval = self.get_time_interval(resource.index.values)
        pysam_w_fname = self._create_pysam_wfile(resource, meta)
        # pylint: disable=E1101
        self[self.PYSAM_WEATHER_TAG] = pysam_w_fname

    def _create_pysam_wfile(self, resource, meta):
        """
        Create PySAM weather input file. PySAM will not accept data on Feb
        29th. For leap years, December 31st is dropped and time steps are
        shifted to relabel Feb 29th as March 1st, March 1st as March 2nd, etc.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.

        Returns
        -------
        fname : string
            Name of weather csv file
        """
        fname = '{}_weather.csv'.format(self._site)
        logger.debug('Creating PySAM weather data file: {}'.format(fname))

        # ------- Process metadata
        m = pd.DataFrame(meta).T
        timezone = m['timezone']
        m['Source'] = 'NSRDB'
        m['Location ID'] = meta.name
        m['City'] = '-'
        m['State'] = m['state'].apply(lambda x: '-' if x == 'None' else x)
        m['Country'] = m['country'].apply(lambda x: '-' if x == 'None' else x)
        m['Latitude'] = m['latitude']
        m['Longitude'] = m['longitude']
        m['Time Zone'] = timezone
        m['Elevation'] = m['elevation']
        m['Local Time Zone'] = timezone
        m['Dew Point Units'] = 'c'
        m['DHI Units'] = 'w/m2'
        m['DNI Units'] = 'w/m2'
        m['Temperature Units'] = 'c'
        m['Pressure Units'] = 'mbar'
        m['Wind Speed'] = 'm/s'
        m = m.drop(['elevation', 'timezone', 'country', 'state', 'county',
                    'urban', 'population', 'landcover', 'latitude',
                    'longitude'], axis=1)
        m.to_csv(fname, index=False, mode='w')

        # --------- Process data
        var_map = {'dni': 'DNI',
                   'dhi': 'DHI',
                   'wind_speed': 'Wind Speed',
                   'air_temperature': 'Temperature',
                   'dew_point': 'Dew Point',
                   'surface_pressure': 'Pressure',
                   }
        resource = resource.rename(mapper=var_map, axis='columns')

        time_index = resource.index
        # Adjust from UTC to local time
        local = np.roll(resource.values, int(timezone * self.time_interval),
                        axis=0)
        resource = pd.DataFrame(local, columns=resource.columns,
                                index=time_index)
        mask = (time_index.month == 2) & (time_index.day == 29)
        time_index = time_index[~mask]

        df = pd.DataFrame(index=time_index)
        df['Year'] = time_index.year
        df['Month'] = time_index.month
        df['Day'] = time_index.day
        df['Hour'] = time_index.hour
        df['Minute'] = time_index.minute
        df = df.join(resource.loc[~mask])

        df.to_csv(fname, index=False, mode='a')

        return fname

    def run_gen_and_econ(self, delete_wfile=True):
        """
        Run SAM generation with possibility for follow on econ analysis.

        Parameters
        ----------
        delete_wfile : bool
            Delete PySAM weather file after processing is complete
        """
        super().run_gen_and_econ()

        pysam_w_fname = self[self.PYSAM_WEATHER_TAG]
        if delete_wfile and os.path.exists(pysam_w_fname):
            os.remove(pysam_w_fname)


class SolarWaterHeat(AbstractSamSolarThermal):
    """
    Solar Water Heater generation
    """
    MODULE = 'solarwaterheat'
    PYSAM = PySamSwh
    PYSAM_WEATHER_TAG = 'solar_resource_file'

    @staticmethod
    def default():
        """Get the executed default pysam swh object.

        Returns
        -------
        PySAM.Swh
        """
        return DefaultSwh.default()


class LinearDirectSteam(AbstractSamSolarThermal):
    """
    Process heat linear Fresnel direct steam generation
    """
    MODULE = 'lineardirectsteam'
    PYSAM = PySamLds
    PYSAM_WEATHER_TAG = 'file_name'

    def cf_mean(self):
        """Calculate mean capacity factor (fractional) from SAM.

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
        """
        net_power = self['annual_field_energy'] \
            - self['annual_thermal_consumption']  # kW-hr
        # q_pb_des is in MW, convert to kW-hr
        name_plate = self['q_pb_des'] * 8760 * 1000

        return net_power / name_plate

    @staticmethod
    def default():
        """Get the executed default pysam linear Fresnel object.

        Returns
        -------
        PySAM.LinearFresnelDsgIph
        """
        return DefaultLinearFresnelDsgIph.default()


class TroughPhysicalHeat(AbstractSamSolarThermal):
    """
    Trough Physical Process Heat generation
    """
    MODULE = 'troughphysicalheat'
    PYSAM = PySamTpph
    PYSAM_WEATHER_TAG = 'file_name'

    def cf_mean(self):
        """Calculate mean capacity factor (fractional) from SAM.

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
        """
        net_power = self['annual_gross_energy'] \
            - self['annual_thermal_consumption']  # kW-hr
        # q_pb_des is in MW, convert to kW-hr
        name_plate = self['q_pb_design'] * 8760 * 1000

        return net_power / name_plate

    @staticmethod
    def default():
        """Get the executed default pysam trough object.

        Returns
        -------
        PySAM.TroughPhysicalProcessHeat
        """
        return DefaultTroughPhysicalProcessHeat.default()


class WindPower(AbstractSamGeneration):
    """Class for Wind generation from SAM
    """
    MODULE = 'windpower'
    PYSAM = PySamWindPower

    def set_resource_data(self, resource, meta):
        """Set WTK resource data arrays.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        """

        # map resource data names to SAM required data names
        var_map = {'speed': 'windspeed',
                   'direction': 'winddirection',
                   'airtemperature': 'temperature',
                   'temp': 'temperature',
                   'surfacepressure': 'pressure',
                   'relativehumidity': 'rh',
                   'humidity': 'rh',
                   }
        lower_case = {k: k.lower().replace(' ', '').replace('_', '')
                      for k in resource.columns}
        resource = resource.rename(mapper=lower_case, axis='columns')
        resource = resource.rename(mapper=var_map, axis='columns')

        data_dict = {}
        var_list = ['temperature', 'pressure', 'windspeed', 'winddirection']
        if 'winddirection' not in resource:
            resource['winddirection'] = 0.0

        time_index = resource.index
        self.time_interval = self.get_time_interval(resource.index.values)

        data_dict['fields'] = [1, 2, 3, 4]
        data_dict['heights'] = 4 * [self.sam_sys_inputs['wind_turbine_hub_ht']]

        if 'rh' in resource:
            # set relative humidity for icing.
            rh = np.roll(self.ensure_res_len(resource['rh'].values,
                                             time_index),
                         int(meta['timezone'] * self.time_interval),
                         axis=0)
            data_dict['rh'] = rh.tolist()

        # must be set as matrix in [temperature, pres, speed, direction] order
        # ensure that resource array length is multiple of 8760
        # roll the truncated resource array to local timezone
        temp = np.roll(self.ensure_res_len(resource[var_list].values,
                                           time_index),
                       int(meta['timezone'] * self.time_interval), axis=0)
        data_dict['data'] = temp.tolist()

        data_dict['lat'] = meta['latitude']
        data_dict['lon'] = meta['longitude']
        data_dict['tz'] = meta['timezone']
        data_dict['elev'] = meta['elevation']

        time_index = self.ensure_res_len(time_index, time_index)
        data_dict['minute'] = time_index.minute
        data_dict['hour'] = time_index.hour
        data_dict['year'] = time_index.year
        data_dict['month'] = time_index.month
        data_dict['day'] = time_index.day

        # add resource data to self.data and clear
        self['wind_resource_data'] = data_dict

    @staticmethod
    def default():
        """Get the executed default pysam WindPower object.

        Returns
        -------
        PySAM.Windpower
        """
        return DefaultWindPower.default()


class MhkWave(AbstractSamGeneration):
    """Class for Wave generation from SAM
    """
    MODULE = 'mhkwave'
    PYSAM = PySamMhkWave

    def set_resource_data(self, resource, meta):
        """Set Hindcast US Wave resource data arrays.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries resource data for a single location with a
            pandas DatetimeIndex. There must be columns for all the required
            variables to run the respective SAM simulation.
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        """

        # map resource data names to SAM required data names
        var_map = {'significantwaveheight': 'significant_wave_height',
                   'waveheight': 'significant_wave_height',
                   'height': 'significant_wave_height',
                   'swh': 'significant_wave_height',
                   'energyperiod': 'energy_period',
                   'waveperiod': 'energy_period',
                   'period': 'energy_period',
                   'ep': 'energy_period',
                   }
        lower_case = {k: k.lower().replace(' ', '').replace('_', '')
                      for k in resource.columns}
        resource = resource.rename(mapper=lower_case, axis='columns')
        resource = resource.rename(mapper=var_map, axis='columns')

        data_dict = {}

        time_index = resource.index
        self.time_interval = self.get_time_interval(resource.index.values)

        # must be set as matrix in [temperature, pres, speed, direction] order
        # ensure that resource array length is multiple of 8760
        # roll the truncated resource array to local timezone
        for var in ['significant_wave_height', 'energy_period']:
            data_dict[var] = np.roll(
                self.ensure_res_len(resource[var].values, time_index),
                int(meta['timezone'] * self.time_interval), axis=0).tolist()

        data_dict['lat'] = meta['latitude']
        data_dict['lon'] = meta['longitude']
        data_dict['tz'] = meta['timezone']

        time_index = self.ensure_res_len(time_index, time_index)
        data_dict['minute'] = time_index.minute
        data_dict['hour'] = time_index.hour
        data_dict['year'] = time_index.year
        data_dict['month'] = time_index.month
        data_dict['day'] = time_index.day

        # add resource data to self.data and clear
        self['wave_resource_data'] = data_dict

    @staticmethod
    def default():
        """Get the executed default PySAM MhkWave object.

        Returns
        -------
        PySAM.MhkWave
        """
        return DefaultMhkWave.default()



# -*- coding: utf-8 -*-
"""
reV data pipeline architecture.
"""
import time
import os
import numpy as np
import logging
from warnings import warn

from reV.config.base_analysis_config import AnalysisConfig
from reV.config.pipeline import PipelineConfig
from reV.pipeline.status import Status
from reV.utilities.exceptions import ExecutionError
from reV.utilities import ModuleName

from rex.utilities.execution import SubprocessManager
from rex.utilities.hpc import SLURM
from rex.utilities.loggers import init_logger
from rex.utilities.utilities import safe_json_load

logger = logging.getLogger(__name__)


class Pipeline:
    """reV pipeline execution framework."""

    CMD_BASE = 'python -m reV.cli -c {fp_config} {command}'
    COMMANDS = ModuleName.all_names()
    RETURN_CODES = {0: 'successful',
                    1: 'running',
                    2: 'failed',
                    3: 'complete'}

    def __init__(self, pipeline, monitor=True, verbose=False):
        """
        Parameters
        ----------
        pipeline : str | dict
            Pipeline config file path or dictionary.
        monitor : bool
            Flag to perform continuous monitoring of the pipeline.
        verbose : bool
            Flag to submit pipeline steps with -v flag for debug logging
        """
        self.monitor = monitor
        self.verbose = verbose
        self._config = PipelineConfig(pipeline)
        self._run_list = self._config.pipeline
        self._init_status()

        # init logger for pipeline module if requested in input config
        if 'logging' in self._config:
            init_logger('reV', **self._config.logging)

    def _init_status(self):
        """Initialize the status json in the output directory."""

        status = self._get_status_obj()

        for i, step in enumerate(self._run_list):

            for module in step.keys():
                module_dict = {module: {'pipeline_index': i}}
                status.data = status.update_dict(status.data, module_dict)

        status._dump()

    def _cancel_all_jobs(self):
        """Cancel all jobs in this pipeline via SLURM scancel."""
        status = self._get_status_obj()
        s = SLURM()
        for job_id in status.job_ids:
            s.scancel(job_id)
        logger.info('Pipeline job "{}" cancelled.'.format(self._config.name))

    def _main(self):
        """Iterate through run list submitting steps while monitoring status"""

        i = 0

        for i, step in enumerate(self._run_list):
            return_code = self._check_step_completed(i)

            if return_code == 0:
                logger.debug('Successful: "{}".'.format(list(step.keys())[0]))
            else:
                return_code = 1
                self._submit_step(i)

                # do not enter while loop for continuous monitoring
                if not self.monitor:
                    break

                time.sleep(1)
                while return_code == 1 and self.monitor:
                    time.sleep(5)
                    return_code = self._check_step_completed(i)

                    if return_code == 2:
                        module, f_config = self._get_command_config(i)
                        raise ExecutionError('Pipeline failed at step '
                                             '{} "{}" {}'
                                             .format(i, module, f_config))

        if i + 1 == len(self._run_list) and return_code == 0:
            logger.info('Pipeline job "{}" is complete.'
                        .format(self._config.name))
            logger.debug('Output directory is: "{}"'
                         .format(self._config.dirout))

    def _submit_step(self, i):
        """Submit a step in the pipeline.

        Parameters
        ----------
        i : int
            Step index in the pipeline run list.
        """

        command, f_config = self._get_command_config(i)
        cmd = self._get_cmd(command, f_config, verbose=self.verbose)

        logger.info('Pipeline submitting: "{}" for job "{}"'
                    .format(command, self._config.name))
        logger.debug('Pipeline submitting subprocess call:\n\t"{}"'
                     .format(cmd))

        try:
            stderr = SubprocessManager.submit(cmd)[1]
        except OSError as e:
            logger.exception('Pipeline subprocess submission returned an '
                             'error: \n{}'.format(e))
            raise e

        if stderr:
            logger.warning('Subprocess received stderr: \n{}'.format(stderr))

    def _check_step_completed(self, i):
        """Check if a pipeline step has been completed.

        Parameters
        ----------
        i : int
            Step index in the pipeline run list.

        Returns
        -------
        return_code : int
            Pipeline step return code.
        """

        module, _ = self._get_command_config(i)
        status = self._get_status_obj()
        submitted = self._check_jobs_submitted(status, module)
        if not submitted:
            return_code = 1
        else:
            return_code = self._get_module_return_code(status, module)

        return return_code

    @staticmethod
    def _check_jobs_submitted(status, module):
        """Check whether jobs have been submitted for a given module.

        Parameters
        ----------
        status : reV.pipeline.status.Status
            reV job status object.
        module : str
            reV module.

        Returns
        -------
        submitted : bool
            Boolean check to see if jobs have been submitted for the module arg
        """

        submitted = False
        if module in status.data:
            jobs = status.data[module]
            for job in jobs.keys():
                if job != 'pipeline_index':
                    submitted = True
                    break
        return submitted

    @staticmethod
    def _get_config_obj(f_config):
        """Get an analysis config object form a config json file.

        Parameters
        ----------
        f_config : str
            File path for config.

        Returns
        -------
        config_obj : AnalysisConfig
            reV analysis config object.
        """

        config_dict = safe_json_load(f_config)
        return AnalysisConfig(config_dict, check_keys=False)

    def _get_status_obj(self):
        """Get a reV pipeline status object.

        Returns
        -------
        status : reV.pipeline.status.Status
            reV job status object.
        """

        status = Status(self._config.dirout, name=self._config.name,
                        hardware=self._config.hardware)
        return status

    def _get_module_return_code(self, status, module):
        """Get a return code for a full module based on a status object.

        Parameters
        ----------
        status : reV.pipeline.status.Status
            reV job status object.
        module : str
            reV module.

        Returns
        -------
        return_code : int
            Pipeline step return code (for the full module in the pipeline
            step).
        """

        # initialize return code array
        arr = []
        check_failed = False

        if module not in status.data:
            # assume running
            arr = [1]
        else:
            for job_name in status.data[module].keys():
                if job_name != 'pipeline_index':

                    # update the job status and get the status string
                    status._update_job_status(module, job_name)
                    js = status.data[module][job_name]['job_status']

                    if js == 'successful':
                        arr.append(0)
                    elif js == 'failed':
                        arr.append(2)
                        check_failed = True
                    elif js is None:
                        arr.append(3)
                    else:
                        arr.append(1)

            status._dump()

        return_code = self._parse_code_array(arr)

        status = self.RETURN_CODES[return_code]
        fail_str = ''
        if check_failed and status != 'failed':
            fail_str = ', but some jobs have failed'
        logger.info('Module "{}" for job "{}" is {}{}.'
                    .format(module, self._config.name, status, fail_str))

        return return_code

    @staticmethod
    def _parse_code_array(arr):
        """Parse array of return codes to get single return code for module.

        Parameters
        ----------
        arr : list | np.ndarray
            List or array of integer return codes.

        Returns
        -------
        return_code : int
            Single return code for the module represented by the input array of
            return codes.
        """

        # check to see if all have completed, or any have failed
        check_success = all(np.array(arr) == 0)
        check_complete = all(np.array(arr) != 1)
        check_failed = any(np.array(arr) == 2)

        # only return success if all have succeeded.
        if check_success:
            return_code = 0
        # Only return failed when all have finished.
        elif check_complete & check_failed:
            return_code = 2
        # only return complete when all have completed
        # (but some should have succeeded or failed)
        elif check_complete:
            return_code = 3
        # otherwise, jobs are still running
        else:
            return_code = 1

        return return_code

    def _get_command_config(self, i):
        """Get the (command, config) key pair.

        Parameters
        ----------
        i : int
            Step index in the pipeline run list.

        Returns
        -------
        key_pair : list
            Two-entry list containing [command, config_file].
        """
        key_pair = list(self._run_list[i].items())[0]
        return key_pair

    @classmethod
    def _get_cmd(cls, command, f_config, verbose=False):
        """Get the python cli call string based on the command and config arg.

        Parameters
        ----------
        command : str
            reV cli command which should be a reV module.
        f_config : str
            File path for the config file corresponding to the command.
        verbose : bool
            Flag to submit pipeline steps with -v flag for debug logging

        Returns
        -------
        cmd : str
            Python reV CLI call string.
        """
        if command not in cls.COMMANDS:
            raise KeyError("Could not recognize command {!r}. "
                           "Available commands are: {!r}"
                           .format(command, cls.COMMANDS)) from None
        cmd = cls.CMD_BASE.format(fp_config=f_config, command=command)
        if verbose:
            cmd += ' -v'

        return cmd

    @staticmethod
    def _get_module_status(status, i):
        """Get the status dict for the module with the given pipeline index.

        Parameters
        ----------
        status : reV.pipeline.status.Status
            reV job status object.
        i : int
            pipeline index of desired module.

        Returns
        -------
        out : dict
            Status dictionary for the module with pipeline index i.
        """

        # iterate through modules and find the one that was run previously
        for module_status in status.data.values():
            i_current = module_status.get('pipeline_index', -99)
            if str(i) == str(i_current):
                out = module_status
                break

        return out

    @staticmethod
    def _get_job_status(module_status, option='all'):
        """Get a job status dict from the module status dict.

        Parameters
        ----------
        module_status : dict
            Status dictionary for a full reV module containing one or more
            job status dict.
        option : str
            Option to retrieve one or many jobs from the module status dict.

        Returns
        -------
        out : dict | list
            Job status(es).
        """

        # find the preceding job (1st is used, should be one job in most cases)
        if option == 'first':
            for job, job_status in module_status.items():
                if job != 'pipeline_index':
                    out = job_status
                    break
        elif option == 'all':
            out = []
            for job, job_status in module_status.items():
                if job != 'pipeline_index':
                    out.append(job_status)
        else:
            raise KeyError('Did not recognize pipeline job status request '
                           'for "{}"'.format(option))
        return out

    @classmethod
    def parse_previous(cls, status_dir, module, target='fpath',
                       target_module=None):
        """Parse output file paths from the previous pipeline step.

        Parameters
        ----------
        status_dir : str
            Directory containing the status file to parse.
        module : str
            Current module (i.e. current pipeline step).
        target : str
            Parsing target of previous module.
        target_module : str | None
            Optional name of module to pull target data from.

        Returns
        -------
        out : list
            Arguments parsed from the status file in status_dir from
            the module preceding the input module arg.
        """

        status = Status(status_dir)
        msg = ('Could not parse data regarding "{}" from status file in '
               '"{}".'.format(module, status_dir))
        if module in status.data:
            if 'pipeline_index' in status.data[module]:
                msg = None
        if msg:
            raise KeyError(msg)

        i1 = int(status.data[module]['pipeline_index'])
        i0 = i1 - 1

        if i0 < 0:
            i0 = 0
            warn('Module "{0}" is attempting to parse a previous pipeline '
                 'step, but it appears to be the first step. Attempting to '
                 'parse data from {0}.'.format(module))

        if target_module is None:
            module_status = cls._get_module_status(status, i0)
            job_statuses = cls._get_job_status(module_status)
        else:
            if target_module not in status.data:
                raise KeyError('Target module "{}" not found in pipeline '
                               'status dictionary.'.format(target_module))
            else:
                module_status = status.data[target_module]
                job_statuses = cls._get_job_status(module_status)

        out = []
        if target == 'fpath':
            for status in job_statuses:
                out.append(os.path.join(status['dirout'], status['fout']))
        else:
            for status in job_statuses:
                out.append(status[target])

        return out

    @classmethod
    def cancel_all(cls, pipeline):
        """Cancel all jobs via SLURM scancel corresponding to pipeline.

        Parameters
        ----------
        pipeline : str | dict
            Pipeline config file path or dictionary.
        """

        pipe = cls(pipeline)
        pipe._cancel_all_jobs()

    @classmethod
    def run(cls, pipeline, monitor=True, verbose=False):
        """Run the reV pipeline.

        Parameters
        ----------
        pipeline : str | dict
            Pipeline config file path or dictionary.
        monitor : bool
            Flag to perform continuous monitoring of the pipeline.
        verbose : bool
            Flag to submit pipeline steps with -v flag for debug logging
        """

        pipe = cls(pipeline, monitor=monitor, verbose=verbose)
        pipe._main()



# -*- coding: utf-8 -*-
"""
reV Configuration for Execution Options
"""
import logging

from reV.config.base_config import BaseConfig
from reV.utilities.exceptions import ConfigError

logger = logging.getLogger(__name__)


class BaseExecutionConfig(BaseConfig):
    """Base class to handle execution configuration"""

    def __init__(self, config_dict):
        """
        Parameters
        ----------
        config : str | dict
            File path to config json (str), serialized json object (str),
            or dictionary with pre-extracted config.
        """

        self._default_option = 'local'
        self._default_nodes = 1
        self._default_mem_util_lim = 0.4
        super().__init__(config_dict)

    @property
    def option(self):
        """Get the hardware run option.

        Returns
        -------
        option : str
            Execution control option, e.g. local, peregrine, eagle...
        """
        return str(self.get('option', self._default_option)).lower()

    @property
    def nodes(self):
        """Get the number of nodes property.

        Returns
        -------
        nodes : int
            Number of available nodes. Default is 1 node.
        """
        return int(self.get('nodes', self._default_nodes))

    @property
    def max_workers(self):
        """Get the max_workers property (1 runs in serial, None is all workers)

        Returns
        -------
        max_workers : int | None
            Processes per node. Default is None max_workers (all available).
        """
        return self.get('max_workers', None)

    @property
    def sites_per_worker(self):
        """Get the number of sites to run per worker.

        Returns
        -------
        sites_per_worker : int | None
            Number of sites to run per worker in a parallel scheme.
        """
        return self.get('sites_per_worker', None)

    @property
    def memory_utilization_limit(self):
        """Get the node memory utilization limit property. Key in the config
        json is "memory_utilization_limit".

        Returns
        -------
        mem_util_lim : float
            Memory utilization limit (fractional). Key in the config json is
            "memory_utilization_limit".
        """
        mem_util_lim = self.get('memory_utilization_limit',
                                self._default_mem_util_lim)

        return mem_util_lim

    @property
    def sh_script(self):
        """Get the "sh_script" entry which is a string that contains extra
        shell script commands to run before the reV commands.

        Returns
        -------
        str
        """
        return self.get('sh_script', '')


class HPCConfig(BaseExecutionConfig):
    """Class to handle HPC configuration inputs."""

    @property
    def allocation(self):
        """Get the HPC allocation property.

        Returns
        -------
        hpc_alloc : str
            Name of the HPC allocation account for the specified job.
        """

        return self.get('allocation', None)

    @property
    def feature(self):
        """Get feature request str.

        Returns
        -------
        feature : str | NoneType
            Feature request string.

            For EAGLE, a full additional flag.
            Config should look like:
                "feature": "--qos=high"
                "feature": "--depend=[state:job_id]"
        """
        return self.get('feature', None)

    @property
    def module(self):
        """
        Get module to load if given

        Returns
        -------
        module : str
            Module to load on node
        """
        return self.get('module', None)

    @property
    def conda_env(self):
        """
        Get conda environment to activate

        Returns
        -------
        conda_env : str
            Conda environment to activate
        """
        return self.get('conda_env', None)


class SlurmConfig(HPCConfig):
    """Class to handle SLURM (Eagle) configuration inputs."""

    def _preflight(self):
        """Run a preflight check on the config."""
        if self.option == 'eagle':
            if self.allocation is None:
                msg = 'Eagle execution config must have an "allocation" input'
                logger.error(msg)
                raise ConfigError(msg)
            if self.walltime is None:
                msg = 'Eagle execution config must have an "walltime" input'
                logger.error(msg)
                raise ConfigError(msg)

        super()._preflight()

    @property
    def memory(self):
        """Get the requested Eagle node "memory" value in GB or can be None.

        Returns
        -------
        _hpc_node_mem : int | None
            Requested node memory in GB.
        """
        return self.get('memory', None)

    @property
    def walltime(self):
        """Get the requested Eagle node "walltime" value.

        Returns
        -------
        _hpc_walltime : int
            Requested single node job time in hours.
        """
        return self.get('walltime', None)



# -*- coding: utf-8 -*-
# pylint: disable=all
"""
reV Representative Profiles command line interface (cli).
"""
import os
import click
import logging
import pprint
import time

from reV.config.hybrids_config import HybridsConfig
from reV.pipeline.status import Status
from reV.hybrids.hybrids import Hybridization, SOLAR_PREFIX, WIND_PREFIX
from reV.utilities import ModuleName
from reV import __version__

from rex.utilities.hpc import SLURM
from rex.utilities.cli_dtypes import STR, INT, FLOATLIST
from rex.utilities.loggers import init_mult
from rex.utilities.utilities import get_class_properties, dict_str_load

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default=os.path.basename(os.getcwd()),
              type=STR, show_default=True, help='reV Hybrids job name.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """reV Hybridization Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid Hybrids config keys
    """
    click.echo(', '.join(get_class_properties(HybridsConfig)))


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='reV hybrid profiles configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run reV hybridization from a config file."""
    # Instantiate the config object
    config = HybridsConfig(config_file)

    # take name from config
    name = ctx.obj['NAME'] = config.name

    # Enforce verbosity if logging level is specified in the config
    if config.log_level == logging.DEBUG:
        verbose = True

    # initialize loggers
    init_mult(name, config.log_directory, modules=[__name__, 'reV', 'rex'],
              verbose=verbose)

    # Initial log statements
    logger.info('Running reV hybridization from config '
                'file: "{}"'.format(config_file))
    logger.info('Target output directory: "{}"'.format(config.dirout))
    logger.info('Target logging directory: "{}"'.format(config.log_directory))
    logger.debug('The full configuration input is as follows:\n{}'
                 .format(pprint.pformat(config, indent=4)))

    names, solar_fpaths, wind_fpaths = _get_paths_from_config(config, name)

    for name, solar_fpath, wind_fpath in zip(names, solar_fpaths, wind_fpaths):

        ctx.obj['NAME'] = name
        if config.execution_control.option == 'local':
            status = Status.retrieve_job_status(config.dirout,
                                                module=ModuleName.HYBRIDS,
                                                job_name=name)

            if status != 'successful':
                Status.add_job(
                    config.dirout, module=ModuleName.HYBRIDS,
                    job_name=name, replace=True,
                    job_attrs={'hardware': 'local',
                               'fout': '{}.h5'.format(name),
                               'dirout': config.dirout})

                ctx.invoke(direct,
                           solar_fpath=solar_fpath,
                           wind_fpath=wind_fpath,
                           allow_solar_only=config.allow_solar_only,
                           allow_wind_only=config.allow_wind_only,
                           fillna=config.fillna,
                           limits=config.limits,
                           ratio_bounds=config.ratio_bounds,
                           ratio=config.ratio,
                           out_dir=config.dirout,
                           log_dir=config.log_directory,
                           verbose=verbose)

        elif config.execution_control.option in ('eagle', 'slurm'):
            ctx.obj['SOLAR_FPATH'] = solar_fpath
            ctx.obj['WIND_FPATH'] = wind_fpath
            ctx.obj['ALLOW_SOLAR_ONLY'] = config.allow_solar_only
            ctx.obj['ALLOW_WIND_ONLY'] = config.allow_wind_only
            ctx.obj['FILLNA'] = config.fillna
            ctx.obj['LIMITS'] = config.limits
            ctx.obj['RATIO_BOUNDS'] = config.ratio_bounds
            ctx.obj['RATIO'] = config.ratio
            ctx.obj['OUT_DIR'] = config.dirout
            ctx.obj['LOG_DIR'] = config.log_directory
            ctx.obj['VERBOSE'] = verbose

            ctx.invoke(slurm,
                       alloc=config.execution_control.allocation,
                       memory=config.execution_control.memory,
                       walltime=config.execution_control.walltime,
                       feature=config.execution_control.feature,
                       conda_env=config.execution_control.conda_env,
                       module=config.execution_control.module,
                       sh_script=config.execution_control.sh_script)


def _get_paths_from_config(config, name):
    """Pair solar and wind files and corresponding process names. """

    solar_glob_paths, wind_glob_paths = config.solar_fpath, config.wind_fpath
    all_years = set(solar_glob_paths) | set(wind_glob_paths)
    common_years = set(solar_glob_paths) & set(wind_glob_paths)
    if not all_years:
        msg = "No files found that match the input: {!r} and/or {!r}"
        e = msg.format(config['solar_fpath'], config['wind_fpath'])
        logger.error(e)
        raise RuntimeError(e)

    solar_fpaths = []
    wind_fpaths = []
    names = []
    for year in all_years:
        if year not in common_years:
            msg = ("No corresponding {} file found for {} input file "
                   "(year: '{}'): {!r}. No hybridization performed for "
                   "this input!")
            resources = (['solar', 'wind'] if year not in solar_glob_paths else
                         ['wind', 'solar'])
            paths = (solar_glob_paths.get(year, [])
                     + wind_glob_paths.get(year, []))
            w = msg.format(*resources, paths, year)
            logger.warning(w)
            RuntimeWarning(w)
            continue

        for fpaths in (solar_glob_paths, wind_glob_paths):
            if len(fpaths[year]) > 1:
                msg = ("Ambiguous number of files found for year '{}': {!r} "
                       "Please ensure there is only one input file per year. "
                       "No hybridization performed for this input!")
                w = msg.format(year, fpaths[year])
                logger.warning(w)
                RuntimeWarning(w)
                break
        else:
            solar_fpaths += solar_glob_paths[year]
            wind_fpaths += wind_glob_paths[year]
            names += ["{}_{}".format(name, year) if year is not None else name]

    return names, solar_fpaths, wind_fpaths


@main.group(invoke_without_command=True)
@click.option('--solar_fpath', '-s', type=click.Path(exists=True),
              required=True, help='Filepath to solar rep profile file.')
@click.option('--wind_fpath', '-w', type=click.Path(exists=True),
              required=True, help='Filepath to wind rep profile file.')
@click.option('-so', '--allow_solar_only', is_flag=True,
              help='Flag to to allow SC points with only solar capcity '
              '(no wind) in hybrid profiles.')
@click.option('-wo', '--allow_wind_only', is_flag=True,
              help='Flag to to allow SC points with only wind capcity '
              '(no solard) in hybrid profiles.')
@click.option('--fillna', '-fna', type=STR, default=None,
              show_default=True,
              help=('String representation of a dictionary of fill values '
                    'for merged columns "{{\'col_name\': val}}" where '
                    '\'col_name\' is the name of a merged column '
                    '(prefixed by {!r} or {!r}) and "val" is the values '
                    'used to fill any n.a. values resulting from a merge '
                    '(e.g. wind column values for a row consisting purely '
                    'of solar capacity).'
                    .format(SOLAR_PREFIX, WIND_PREFIX)))
@click.option('--limits', '-l', type=STR, default=None,
              show_default=True,
              help='String representation of a dictionary specifying a '
                   'mapping of "{{\'colum_name\': max_value}}" representing '
                   'the upper limit (maximum value) for the values of a '
                   'column in the merged meta. For example, '
                   '`--limits "{{\'solar_capacity\': 100}}"` '
                   'would limit all the values of the solar capacity in the '
                   'merged meta to a maximum value of 100. This limit is '
                   'applied *BEFORE* ratio calculations. The names of '
                   'the columns should match the column names in the merged '
                   'meta, so they are likely prefixed by {!r} or {!r}. '
                   'By default, None (no limits applied).'
                   .format(SOLAR_PREFIX, WIND_PREFIX))
@click.option('--ratio_bounds', '-rb', type=FLOATLIST, default=None,
              show_default=True,
              help='List of two floats representing the lower and upper '
                   'bounds on the ratio of the two columns specified with the '
                   '`--ratio` input, which is set to "solar_capacity/'
                   'wind_capacity" by default. The same value can be used as '
                   'the upper and lower bound to represent a single ratio. '
                   'For example, `--ratio_bounds=[0.5, 1.5]` would adjust the '
                   'values of both of the `--ratio` columns such that their '
                   'ratio is always between half and double (e.g., no value '
                   'would be more than double the other). To specify a single '
                   'ratio value, use the same value as the upper and lower '
                   'bound. For example, `--ratio_bounds=(1, 1)` would adjust '
                   'the values of both of the `ratio` columns such that their '
                   'ratio is always equal.')
@click.option('--ratio', '-r', type=STR,
              default='solar_capacity/wind_capacity',
              show_default=True,
              help='String representation of the ratio calculation used to '
                   'calculate the ratio that is limited by the `ratio_bounds`'
                   'input. This string must be in the form '
                   '"numerator_column_name/denominator_column_name". For '
                   'example, `--ratio \'solar_capacity/wind_capacity\'` would '
                   'limit the ratio of the solar to wind capacities as '
                   'specified by the `ratio_bounds` input. If `ratio_bounds` '
                   'is not specified, this input does nothing. The names of '
                   'the columns should match the column names in the merged '
                   'meta, so they are likely prefixed by {!r} or {!r}.'
                   .format(SOLAR_PREFIX, WIND_PREFIX))
@click.option('--out_dir', '-od', type=STR, default='./',
              show_default=True,
              help='Directory to save rep profile output h5.')
@click.option('--log_dir', '-ld', type=STR, default='./logs/',
              show_default=True,
              help='Directory to save rep profile logs.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def direct(ctx, solar_fpath, wind_fpath, allow_solar_only, allow_wind_only,
           fillna, limits, ratio_bounds, ratio, out_dir, log_dir, verbose):
    """reV hybridization CLI."""
    name = ctx.obj['NAME']
    ctx.obj['SOLAR_FPATH'] = solar_fpath
    ctx.obj['WIND_FPATH'] = wind_fpath
    ctx.obj['ALLOW_SOLAR_ONLY'] = allow_solar_only
    ctx.obj['ALLOW_WIND_ONLY'] = allow_wind_only
    ctx.obj['FILLNA'] = fillna
    ctx.obj['LIMITS'] = limits
    ctx.obj['RATIO_BOUNDS'] = ratio_bounds
    ctx.obj['RATIO'] = ratio
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['LOG_DIR'] = log_dir
    ctx.obj['VERBOSE'] = verbose

    if ctx.invoked_subcommand is None:
        t0 = time.time()
        init_mult(name, log_dir, modules=[__name__, 'reV', 'rex'],
                  verbose=verbose)

        fn_out = '{}.h5'.format(name)
        fout = os.path.join(out_dir, fn_out)

        if isinstance(fillna, str):
            fillna = dict_str_load(fillna)

        if isinstance(limits, str):
            limits = dict_str_load(limits)

        try:
            Hybridization(
                solar_fpath, wind_fpath,
                allow_solar_only=allow_solar_only,
                allow_wind_only=allow_wind_only,
                fillna=fillna, limits=limits,
                ratio_bounds=ratio_bounds,
                ratio=ratio
            ).run_all(fout=fout)

        except Exception as e:
            msg = ('Hybridization of rep profiles failed. Received the '
                   'following error:\n{}'.format(e))
            logger.exception(msg)
            raise RuntimeError(msg) from e

        runtime = (time.time() - t0) / 60
        logger.info('reV hybrid profiles complete. '
                    'Time elapsed: {:.2f} min. Target output dir: {}'
                    .format(runtime, out_dir))

        status = {'dirout': out_dir, 'fout': fn_out,
                  'job_status': 'successful',
                  'runtime': runtime,
                  'finput': [solar_fpath, wind_fpath]}
        Status.make_job_file(out_dir, ModuleName.HYBRIDS, name, status)


def get_node_cmd(ctx):
    """Get a CLI call command for the hybrids cli."""

    name = ctx.obj['NAME']
    solar_fpath = ctx.obj['SOLAR_FPATH']
    wind_fpath = ctx.obj['WIND_FPATH']
    allow_solar_only = ctx.obj['ALLOW_SOLAR_ONLY']
    allow_wind_only = ctx.obj['ALLOW_WIND_ONLY']
    fillna = ctx.obj['FILLNA']
    limits = ctx.obj['LIMITS']
    ratio_bounds = ctx.obj['RATIO_BOUNDS']
    ratio = ctx.obj['RATIO']
    out_dir = ctx.obj['OUT_DIR']
    log_dir = ctx.obj['LOG_DIR']
    verbose = ctx.obj['VERBOSE']

    args = ['-s {}'.format(SLURM.s(solar_fpath)),
            '-w {}'.format(SLURM.s(wind_fpath)),
            '-fna {}'.format(SLURM.s(fillna)),
            '-l {}'.format(SLURM.s(limits)),
            '-od {}'.format(SLURM.s(out_dir)),
            '-ld {}'.format(SLURM.s(log_dir)),
            ]

    if ratio_bounds is not None:
        args.append('-rb {}'.format(SLURM.s(ratio_bounds)))

    if ratio is not None:
        args.append('-r {}'.format(SLURM.s(ratio)))

    if allow_solar_only:
        args.append('-so')

    if allow_wind_only:
        args.append('-wo')

    if verbose:
        args.append('-v')

    cmd = ('python -m reV.hybrids.cli_hybrids -n {} direct {}'
           .format(SLURM.s(name), ' '.join(args)))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@direct.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='SLURM allocation account name.')
@click.option('--memory', '-mem', default=None, type=INT,
              show_default=True,
              help='SLURM node memory request in GB. Default is None')
@click.option('--walltime', '-wt', default=1.0, type=float,
              show_default=True,
              help='SLURM walltime request in hours for single year '
              'rep_profiles run. Default is 1.0')
@click.option('--feature', '-l', default=None, type=STR,
              show_default=True,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--conda_env', '-env', default=None, type=STR,
              show_default=True,
              help='Conda env to activate')
@click.option('--module', '-mod', default=None, type=STR,
              show_default=True,
              help='Module to load')
@click.option('--stdout_path', '-sout', default=None, type=STR,
              show_default=True,
              help='Subprocess standard output path. Default is in out_dir.')
@click.option('--sh_script', '-sh', default=None, type=STR,
              show_default=True,
              help='Extra shell script commands to run before the reV call.')
@click.pass_context
def slurm(ctx, alloc, memory, walltime, feature, conda_env, module,
          stdout_path, sh_script):
    """slurm (Eagle) submission tool for reV representative profiles."""

    name = ctx.obj['NAME']
    out_dir = ctx.obj['OUT_DIR']
    log_dir = ctx.obj['LOG_DIR']

    if stdout_path is None:
        stdout_path = os.path.join(log_dir, 'stdout/')

    cmd = get_node_cmd(ctx)

    if sh_script:
        cmd = sh_script + '\n' + cmd

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(out_dir, module=ModuleName.HYBRIDS,
                                        job_name=name, hardware='eagle',
                                        subprocess_manager=slurm_manager)

    msg = 'Hybrids CLI failed to submit jobs!'
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'
               .format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running reV SC hybridization on SLURM with '
                    'node name "{}"'.format(name))
        out = slurm_manager.sbatch(cmd, alloc=alloc, memory=memory,
                                   walltime=walltime, feature=feature,
                                   name=name, stdout_path=stdout_path,
                                   conda_env=conda_env, module=module)[0]
        if out:
            msg = ('Kicked off reV hybridization job "{}" '
                   '(SLURM jobid #{}).'
                   .format(name, out))
        Status.add_job(
            out_dir, module=ModuleName.HYBRIDS, job_name=name, replace=True,
            job_attrs={'job_id': out, 'hardware': 'eagle',
                       'fout': '{}.h5'.format(name), 'dirout': out_dir})

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV Hybridization CLI.')
        raise



# -*- coding: utf-8 -*-
"""
reV-NRWAL config.

@author: gbuster
"""
from glob import glob
import logging

from reV.config.base_analysis_config import AnalysisConfig
from reV.pipeline.pipeline import Pipeline
from reV.utilities import ModuleName

logger = logging.getLogger(__name__)


class RevNrwalConfig(AnalysisConfig):
    """Offshore wind aggregation config."""

    NAME = ModuleName.NRWAL
    REQUIREMENTS = ('gen_fpath', 'site_data', 'sam_files', 'nrwal_configs',
                    'output_request')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            File path to config json (str), serialized json object (str),
            or dictionary with pre-extracted config.
        """
        super().__init__(config)
        self._gen_fpath = self._parse_gen_fpath()

    def _parse_gen_fpath(self):
        """Get one or more generation data filepaths

        Returns
        -------
        list | str
        """
        fpaths = self['gen_fpath']
        if fpaths == 'PIPELINE':
            fpaths = Pipeline.parse_previous(
                self.dirout, module=ModuleName.NRWAL, target='fpath'
            )

        if isinstance(fpaths, str) and '*' in fpaths:
            fpaths = glob(fpaths)
            if not any(fpaths):
                msg = ('Could not find any file paths for '
                       'gen_fpath glob pattern.')
                logger.error(msg)
                raise RuntimeError(msg)

        if len(fpaths) == 1:
            fpaths = fpaths[0]

        return fpaths

    @property
    def gen_fpath(self):
        """Base generation fpath(s) used as input data files. Anything in the
        output_request is added and/or manipulated in this file(s)."""
        return self._gen_fpath

    @property
    def site_data(self):
        """Get the site-specific spatial data filepath"""
        return self['site_data']

    @property
    def sam_files(self):
        """Get the sam files dict"""
        return self['sam_files']

    @property
    def nrwal_configs(self):
        """Get the nrwal configs dict"""
        return self['nrwal_configs']

    @property
    def output_request(self):
        """Get keys from the nrwal configs to pass through as new datasets in
        the reV output h5. If you want to manipulate a dset like cf_mean from
        gen_fpath and include it in the output_request, you should set
        save_raw=True and then in the NRWAL equations use cf_mean_raw as the
        input and then define cf_mean as the manipulated data that will be
        included in the output_request.
        """
        return self['output_request']

    @property
    def save_raw(self):
        """Get the flag to save raw datasets in gen_fpath in the case that they
        are manipulated then requested in output_request. Optional, default is
        True."""
        return bool(self.get('save_raw', True))

    @property
    def meta_gid_col(self):
        """The column in the gen_fpath meta data that corresponds to the "gid"
        column in the site_data input. Optional, default is "gid".
        """
        return self.get('meta_gid_col', 'gid')

    @property
    def site_meta_cols(self):
        """Column labels from site_data to pass through to the output meta
        data. Optional, default is None."""
        return self.get('site_meta_cols', None)



# -*- coding: utf-8 -*-
"""
Compute and plot summary data
"""
import logging
import numpy as np
import os
import pandas as pd
import plotting as mplt
import plotly.express as px

from rex import Resource
from rex.utilities import SpawnProcessPool, parse_table

logger = logging.getLogger(__name__)


class SummarizeH5:
    """
    reV Summary data for QA/QC
    """
    def __init__(self, h5_file, group=None):
        """
        Parameters
        ----------
        h5_file : str
            Path to .h5 file to summarize data from
        group : str, optional
            Group within h5_file to summarize datasets for, by default None
        """
        logger.info('QAQC Summarize initializing on: {}'.format(h5_file))
        self._h5_file = h5_file
        self._group = group

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self.h5_file)

        return msg

    @property
    def h5_file(self):
        """
        .h5 file path

        Returns
        -------
        str
        """
        return self._h5_file

    @staticmethod
    def _compute_sites_summary(h5_file, ds_name, sites=None, group=None):
        """
        Compute summary stats for given sites of given dataset

        Parameters
        ----------
        h5_file : str
            Path to .h5 file to summarize data from
        ds_name : str
            Dataset name of interest
        sites : list | slice, optional
            sites of interest, by default None
        group : str, optional
            Group within h5_file to summarize datasets for, by default None

        Returns
        -------
        sites_summary : pandas.DataFrame
            Summary stats for given sites / dataset
        """
        if sites is None:
            sites = slice(None)

        with Resource(h5_file, group=group) as f:
            sites_meta = f['meta', sites]
            sites_data = f[ds_name, :, sites]

        sites_summary = pd.DataFrame(sites_data, columns=sites_meta.index)
        sites_summary = sites_summary.describe().T.drop(columns=['count'])
        sites_summary['sum'] = sites_data.sum(axis=0)

        return sites_summary

    @staticmethod
    def _compute_ds_summary(h5_file, ds_name, group=None):
        """
        Compute summary statistics for given dataset (assumed to be a vector)

        Parameters
        ----------
        h5_file : str
            Path to .h5 file to summarize data from
        ds_name : str
            Dataset name of interest
        group : str, optional
            Group within h5_file to summarize datasets for, by default None

        Returns
        -------
        ds_summary : pandas.DataFrame
            Summary statistics for dataset
        """
        with Resource(h5_file, group=group) as f:
            ds_data = f[ds_name, :]

        ds_summary = pd.DataFrame(ds_data, columns=[ds_name])
        ds_summary = ds_summary.describe().drop(['count'])
        ds_summary.at['sum'] = ds_data.sum()

        return ds_summary

    def summarize_dset(self, ds_name, process_size=None, max_workers=None,
                       out_path=None):
        """
        Compute dataset summary. If dataset is 2D compute temporal statistics
        for each site

        Parameters
        ----------
        ds_name : str
            Dataset name of interest
        process_size : int, optional
            Number of sites to process at a time, by default None
        max_workers : int, optional
            Number of workers to use in parallel, if 1 run in serial,
            if None use all available cores, by default None
        out_path : str
            File path to save summary to

        Returns
        -------
        summary : pandas.DataFrame
            Summary summary for dataset
        """
        with Resource(self.h5_file, group=self._group) as f:
            ds_shape, _, ds_chunks = f.get_dset_properties(ds_name)

        if len(ds_shape) > 1:
            sites = np.arange(ds_shape[1])
            if max_workers != 1:
                if process_size is None and ds_chunks is not None:
                    process_size = ds_chunks[1]
                if process_size is None:
                    process_size = ds_shape[-1]

                sites = \
                    np.array_split(sites,
                                   int(np.ceil(len(sites) / process_size)))
                loggers = [__name__, 'reV']
                with SpawnProcessPool(max_workers=max_workers,
                                      loggers=loggers) as ex:
                    futures = []
                    for site_slice in sites:
                        futures.append(ex.submit(
                            self._compute_sites_summary,
                            self.h5_file, ds_name, sites=site_slice,
                            group=self._group))

                    summary = [future.result() for future in futures]

                summary = pd.concat(summary)
            else:
                if process_size is None:
                    summary = self._compute_sites_summary(self.h5_file,
                                                          ds_name,
                                                          sites=sites,
                                                          group=self._group)
                else:
                    sites = np.array_split(
                        sites, int(np.ceil(len(sites) / process_size)))

                    summary = []
                    for site_slice in sites:
                        summary.append(self._compute_sites_summary(
                            self.h5_file, ds_name,
                            sites=site_slice,
                            group=self._group))

                    summary = pd.concat(summary)

            summary.index.name = 'gid'

        else:
            summary = self._compute_ds_summary(self.h5_file, ds_name,
                                               group=self._group)

        if out_path is not None:
            summary.to_csv(out_path)

        return summary

    def summarize_means(self, out_path=None):
        """
        Add means datasets to meta data

        Parameters
        ----------
        out_path : str, optional
            Path to .csv file to save update meta data to, by default None

        Returns
        -------
        meta : pandas.DataFrame
            Meta data with means datasets added
        """
        with Resource(self.h5_file, group=self._group) as f:
            meta = f.meta
            if 'gid' not in meta:
                if meta.index.name != 'gid':
                    meta.index.name = 'gid'

                meta = meta.reset_index()

            for ds_name in f.datasets:
                shape, dtype, _ = f.get_dset_properties(ds_name)
                if len(shape) == 1 and np.issubdtype(dtype, np.number):
                    meta[ds_name] = f[ds_name]

        if out_path is not None:
            meta.to_csv(out_path, index=False)

        return meta

    @classmethod
    def run(cls, h5_file, out_dir, group=None, dsets=None,
            process_size=None, max_workers=None):
        """
        Summarize all datasets in h5_file and dump to out_dir

        Parameters
        ----------
        h5_file : str
            Path to .h5 file to summarize data from
        out_dir : str
            Directory to dump summary .csv files to
        group : str, optional
            Group within h5_file to summarize datasets for, by default None
        dsets : str | list, optional
            Datasets to summarize, by default None
        process_size : int, optional
            Number of sites to process at a time, by default None
        max_workers : int, optional
            Number of workers to use when summarizing 2D datasets,
            by default None
        """
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        if dsets is None:
            with Resource(h5_file, group=group) as f:
                dsets = [dset for dset in f.datasets
                         if dset not in ['meta', 'time_index']]
        elif isinstance(dsets, str):
            dsets = [dsets]

        summary = cls(h5_file)
        for ds_name in dsets:
            out_path = os.path.join(out_dir,
                                    "{}_summary.csv".format(ds_name))
            summary.summarize_dset(ds_name, process_size=process_size,
                                   max_workers=max_workers, out_path=out_path)

        out_path = os.path.basename(h5_file).replace('.h5', '_summary.csv')
        out_path = os.path.join(out_dir, out_path)
        summary.summarize_means(out_path=out_path)


class SummarizeSupplyCurve:
    """
    Summarize Supply Curve table
    """
    def __init__(self, sc_table):
        self._sc_table = self._parse_summary(sc_table)

    def __repr__(self):
        msg = "{}".format(self.__class__.__name__)

        return msg

    @property
    def sc_table(self):
        """
        Supply Curve table

        Returns
        -------
        pd.DataFrame
        """
        return self._sc_table

    @staticmethod
    def _parse_summary(summary):
        """
        Extract summary statistics

        Parameters
        ----------
        summary : str | pd.DataFrame
            Path to .csv or .json or DataFrame to parse

        Returns
        -------
        summary : pandas.DataFrame
            DataFrame of summary statistics
        """
        try:
            summary = parse_table(summary)
        except ValueError as ex:
            logger.error(ex)
            raise

        return summary

    def supply_curve_summary(self, columns=None, out_path=None):
        """
        Summarize Supply Curve Table

        Parameters
        ----------
        sc_table : str | pandas.DataFrame
            Supply curve table or .csv containing table
        columns : str | list, optional
            Column(s) to summarize, if None summarize all numeric columns,
            by default None
        out_path : str, optional
            Path to .csv to save summary to, by default None

        Returns
        -------
        sc_summary : pandas.DataFrame
            Summary statistics (mean, stdev, median, min, max, sum) for
            Supply Curve table columns
        """
        sc_table = self.sc_table
        if columns is not None:
            if isinstance(columns, str):
                columns = [columns]

            sc_table = sc_table[columns]

        sc_table = sc_table.select_dtypes(include=np.number)

        sc_summary = []
        sc_stat = sc_table.mean(axis=0)
        sc_stat.name = 'mean'
        sc_summary.append(sc_stat)

        sc_stat = sc_table.std(axis=0)
        sc_stat.name = 'stdev'
        sc_summary.append(sc_stat)

        sc_stat = sc_table.median(axis=0)
        sc_stat.name = 'median'
        sc_summary.append(sc_stat)

        sc_stat = sc_table.min(axis=0)
        sc_stat.name = 'min'
        sc_summary.append(sc_stat)

        sc_stat = sc_table.max(axis=0)
        sc_stat.name = 'max'
        sc_summary.append(sc_stat)

        sc_stat = sc_table.sum(axis=0)
        sc_stat.name = 'sum'
        sc_summary.append(sc_stat)

        sc_summary = pd.concat(sc_summary, axis=1).T

        if out_path is not None:
            sc_summary.to_csv(out_path)

        return sc_summary

    @classmethod
    def run(cls, sc_table, out_dir, columns=None):
        """
        Summarize Supply Curve Table and save to disk

        Parameters
        ----------
        sc_table : str | pandas.DataFrame
            Path to .csv containing Supply Curve table
        out_dir : str
            Directory to dump summary .csv files to
        columns : str | list, optional
            Column(s) to summarize, if None summarize all numeric columns,
            by default None
        """
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        summary = cls(sc_table)
        out_path = os.path.basename(sc_table).replace('.csv', '_summary.csv')
        out_path = os.path.join(out_dir, out_path)
        summary.supply_curve_summary(columns=columns, out_path=out_path)


class PlotBase:
    """
    QA/QC Plotting base class
    """
    def __init__(self, data):
        """
        Parameters
        ----------
        data : str | pandas.DataFrame | ndarray
            data to plot or file containing data to plot
        """
        self._data = data

    def __repr__(self):
        msg = "{}".format(self.__class__.__name__)

        return msg

    @property
    def data(self):
        """
        Data to plot

        Returns
        -------
        pandas.DataFrame | ndarray
        """
        return self._data

    @staticmethod
    def _save_plotly(fig, out_path):
        """
        Save plotly figure to disk

        Parameters
        ----------
        fig : plotly.Figure
            Plotly Figure object
        out_path : str
            File path to save plot to, can be a .html or static image
        """
        if out_path.endswith('.html'):
            fig.write_html(out_path)
        else:
            fig.write_image(out_path)

    @staticmethod
    def _check_value(df, values, scatter=True):
        """
        Check DataFrame for needed columns

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame to check
        values : str | list
            Column(s) to plot
        scatter : bool, optional
            Flag to check for latitude and longitude columns, by default True
        """
        if isinstance(values, str):
            values = [values]

        if scatter:
            values += ['latitude', 'longitude']

        for value in values:
            if value not in df:
                msg = ("{} is not a valid column in summary table:\n{}"
                       .format(value, df))
                logger.error(msg)
                raise ValueError(msg)


class SummaryPlots(PlotBase):
    """
    Plot summary data for QA/QC
    """
    def __init__(self, summary):
        """
        Parameters
        ----------
        summary : str | pandas.DataFrame
            Summary DataFrame or path to summary .csv
        """
        self._data = SummarizeSupplyCurve._parse_summary(summary)

    @property
    def summary(self):
        """
        Summary table

        Returns
        -------
        pandas.DataFrame
        """
        return self._data

    @property
    def columns(self):
        """
        Available columns in summary table

        Returns
        -------
        list
        """
        return list(self.summary.columns)

    def scatter_plot(self, value, cmap='viridis', out_path=None, **kwargs):
        """
        Plot scatter plot of value versus longitude and latitude using
        pandas.plot.scatter

        Parameters
        ----------
        value : str
            Column name to plot as color
        cmap : str, optional
            Matplotlib colormap name, by default 'viridis'
        out_path : str, optional
            File path to save plot to, by default None
        kwargs : dict
            Additional kwargs for plotting.dataframes.df_scatter
        """
        self._check_value(self.summary, value)
        mplt.df_scatter(self.summary, x='longitude', y='latitude', c=value,
                        colormap=cmap, filename=out_path, **kwargs)

    def scatter_plotly(self, value, cmap='Viridis', out_path=None, **kwargs):
        """
        Plot scatter plot of value versus longitude and latitude using
        plotly

        Parameters
        ----------
        value : str
            Column name to plot as color
        cmap : str | px.color, optional
            Continuous color scale to use, by default 'Viridis'
        out_path : str, optional
            File path to save plot to, can be a .html or static image,
            by default None
        kwargs : dict
            Additional kwargs for plotly.express.scatter
        """
        self._check_value(self.summary, value)
        fig = px.scatter(self.summary, x='longitude', y='latitude',
                         color=value, color_continuous_scale=cmap, **kwargs)
        fig.update_layout(font=dict(family="Arial", size=18, color="black"))

        if out_path is not None:
            self._save_plotly(fig, out_path)

        fig.show()

    def _extract_sc_data(self, lcoe='mean_lcoe'):
        """
        Extract supply curve data

        Parameters
        ----------
        lcoe : str, optional
            LCOE value to use for supply curve, by default 'mean_lcoe'

        Returns
        -------
        sc_df : pandas.DataFrame
            Supply curve data
        """
        values = ['capacity', lcoe]
        self._check_value(self.summary, values, scatter=False)
        sc_df = self.summary[values].sort_values(lcoe)
        sc_df['cumulative_capacity'] = sc_df['capacity'].cumsum()

        return sc_df

    def dist_plot(self, value, out_path=None, **kwargs):
        """
        Plot distribution plot of value using seaborn.distplot

        Parameters
        ----------
        value : str
            Column name to plot
        out_path : str, optional
            File path to save plot to, by default None
        kwargs : dict
            Additional kwargs for plotting.dataframes.dist_plot
        """
        self._check_value(self.summary, value, scatter=False)
        series = self.summary[value]
        mplt.dist_plot(series, filename=out_path, **kwargs)

    def dist_plotly(self, value, out_path=None, **kwargs):
        """
        Plot histogram of value using plotly

        Parameters
        ----------
        value : str
            Column name to plot
        out_path : str, optional
            File path to save plot to, by default None
        kwargs : dict
            Additional kwargs for plotly.express.histogram
        """
        self._check_value(self.summary, value, scatter=False)

        fig = px.histogram(self.summary, x=value)

        if out_path is not None:
            self._save_plotly(fig, out_path, **kwargs)

        fig.show()

    @classmethod
    def scatter(cls, summary_csv, out_dir, value, plot_type='plotly',
                cmap='viridis', **kwargs):
        """
        Create scatter plot for given value in summary table and save to
        out_dir

        Parameters
        ----------
        summary_csv : str
            Path to .csv file containing summary table
        out_dir : str
            Output directory to save plots to
        value : str
            Column name to plot as color
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        kwargs : dict
            Additional plotting kwargs
        """
        splt = cls(summary_csv)
        if plot_type == 'plot':
            out_path = os.path.basename(summary_csv).replace('.csv', '.png')
            out_path = os.path.join(out_dir, out_path)
            splt.scatter_plot(value, cmap=cmap.lower(), out_path=out_path,
                              **kwargs)
        elif plot_type == 'plotly':
            out_path = os.path.basename(summary_csv).replace('.csv', '.html')
            out_path = os.path.join(out_dir, out_path)
            splt.scatter_plotly(value, cmap=cmap.capitalize(),
                                out_path=out_path, **kwargs)
        else:
            msg = ("plot_type must be 'plot' or 'plotly' but {} was given"
                   .format(plot_type))
            logger.error(msg)
            raise ValueError(msg)

    @classmethod
    def scatter_all(cls, summary_csv, out_dir, plot_type='plotly',
                    cmap='viridis', **kwargs):
        """
        Create scatter plot for all summary stats in summary table and save to
        out_dir

        Parameters
        ----------
        summary_csv : str
            Path to .csv file containing summary table
        out_dir : str
            Output directory to save plots to
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        kwargs : dict
            Additional plotting kwargs
        """
        splt = cls(summary_csv)
        splt._data = splt.summary.select_dtypes(include=np.number)
        datasets = [c for c in splt.summary.columns
                    if not c.startswith(('lat', 'lon'))]

        for value in datasets:
            if plot_type == 'plot':
                out_path = '_{}.png'.format(value)
                out_path = \
                    os.path.basename(summary_csv).replace('.csv', out_path)
                out_path = os.path.join(out_dir, out_path)
                splt.scatter_plot(value, cmap=cmap.lower(), out_path=out_path,
                                  **kwargs)
            elif plot_type == 'plotly':
                out_path = '_{}.html'.format(value)
                out_path = \
                    os.path.basename(summary_csv).replace('.csv', out_path)
                out_path = os.path.join(out_dir, out_path)
                splt.scatter_plotly(value, cmap=cmap.capitalize(),
                                    out_path=out_path, **kwargs)
            else:
                msg = ("plot_type must be 'plot' or 'plotly' but {} was given"
                       .format(plot_type))
                logger.error(msg)
                raise ValueError(msg)


class SupplyCurvePlot(PlotBase):
    """
    Plot supply curve data for QA/QC
    """

    def __init__(self, sc_table):
        """
        Parameters
        ----------
        sc_table : str | pandas.DataFrame
            Supply curve table or path to supply curve .csv
        """
        self._data = SummarizeSupplyCurve._parse_summary(sc_table)

    @property
    def sc_table(self):
        """
        Supply curve table

        Returns
        -------
        pandas.DataFrame
        """
        return self._data

    @property
    def columns(self):
        """
        Available columns in supply curve table

        Returns
        -------
        list
        """
        return list(self.sc_table.columns)

    def _extract_sc_data(self, lcoe='mean_lcoe'):
        """
        Extract supply curve data

        Parameters
        ----------
        lcoe : str, optional
            LCOE value to use for supply curve, by default 'mean_lcoe'

        Returns
        -------
        sc_df : pandas.DataFrame
            Supply curve data
        """
        values = ['capacity', lcoe]
        self._check_value(self.sc_table, values, scatter=False)
        sc_df = self.sc_table[values].sort_values(lcoe)
        sc_df['cumulative_capacity'] = sc_df['capacity'].cumsum()

        return sc_df

    def supply_curve_plot(self, lcoe='mean_lcoe', out_path=None, **kwargs):
        """
        Plot supply curve (cumulative capacity vs lcoe) using seaborn.scatter

        Parameters
        ----------
        lcoe : str, optional
            LCOE value to plot, by default 'mean_lcoe'
        out_path : str, optional
            File path to save plot to, by default None
        kwargs : dict
            Additional kwargs for plotting.dataframes.df_scatter
        """
        sc_df = self._extract_sc_data(lcoe=lcoe)
        mplt.df_scatter(sc_df, x='cumulative_capacity', y=lcoe,
                        filename=out_path, **kwargs)

    def supply_curve_plotly(self, lcoe='mean_lcoe', out_path=None, **kwargs):
        """
        Plot supply curve (cumulative capacity vs lcoe) using plotly

        Parameters
        ----------
        lcoe : str, optional
            LCOE value to plot, by default 'mean_lcoe'
        out_path : str, optional
            File path to save plot to, can be a .html or static image,
            by default None
        kwargs : dict
            Additional kwargs for plotly.express.scatter
        """
        sc_df = self._extract_sc_data(lcoe=lcoe)
        fig = px.scatter(sc_df, x='cumulative_capacity', y=lcoe, **kwargs)
        fig.update_layout(font=dict(family="Arial", size=18, color="black"))

        if out_path is not None:
            self._save_plotly(fig, out_path)

        fig.show()

    @classmethod
    def plot(cls, sc_table, out_dir, plot_type='plotly', lcoe='mean_lcoe',
             **kwargs):
        """
        Create supply curve plot from supply curve table using lcoe value
        and save to out_dir

        Parameters
        ----------
        sc_table : str
            Path to .csv file containing Supply Curve table
        out_dir : str
            Output directory to save plots to
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        lcoe : str, optional
            LCOE value to plot, by default 'mean_lcoe'
        kwargs : dict
            Additional plotting kwargs
        """
        splt = cls(sc_table)
        if plot_type == 'plot':
            out_path = os.path.basename(sc_table).replace('.csv', '.png')
            out_path = os.path.join(out_dir, out_path)
            splt.supply_curve_plot(lcoe=lcoe, out_path=out_path, **kwargs)
        elif plot_type == 'plotly':
            out_path = os.path.basename(sc_table).replace('.csv', '.html')
            out_path = os.path.join(out_dir, out_path)
            splt.supply_curve_plotly(lcoe=lcoe, out_path=out_path, **kwargs)
        else:
            msg = ("plot_type must be 'plot' or 'plotly' but {} was given"
                   .format(plot_type))
            logger.error(msg)
            raise ValueError(msg)


class ExclusionsMask(PlotBase):
    """
    Plot Exclusions mask as a heat map data for QA/QC
    """

    def __init__(self, excl_mask):
        """
        Parameters
        ----------
        excl_mask : str | ndarray
            Exclusions mask or path to .npy file containing final mask
        """
        self._data = self._parse_mask(excl_mask)

    @property
    def mask(self):
        """
        Final Exclusions mask

        Returns
        -------
        ndarray
        """
        return self._data

    @staticmethod
    def _parse_mask(excl_mask):
        """
        Load exclusions mask if needed

        Parameters
        ----------
        excl_mask : str | ndarray
            Exclusions mask or path to .npy file containing final mask

        Returns
        -------
        excl_mask : ndarray
            [n, m] array of final exclusion values
        """
        if isinstance(excl_mask, str):
            excl_mask = np.load(excl_mask)
        elif not isinstance(excl_mask, np.ndarray):
            raise ValueError("excl_mask must be a .npy file or an ndarray")

        return excl_mask

    def exclusions_plot(self, cmap='Viridis', plot_step=100, out_path=None,
                        **kwargs):
        """
        Plot exclusions mask as a seaborn heatmap

        Parameters
        ----------
        cmap : str | px.color, optional
            Continuous color scale to use, by default 'Viridis'
        plot_step : int
            Step between points to plot
        out_path : str, optional
            File path to save plot to, can be a .html or static image,
            by default None
        kwargs : dict
            Additional kwargs for plotting.colormaps.heatmap_plot
        """
        mplt.heatmap_plot(self.mask[::plot_step, ::plot_step], cmap=cmap,
                          filename=out_path, **kwargs)

    def exclusions_plotly(self, cmap='Viridis', plot_step=100, out_path=None,
                          **kwargs):
        """
        Plot exclusions mask as a plotly heatmap

        Parameters
        ----------
        cmap : str | px.color, optional
            Continuous color scale to use, by default 'Viridis'
        plot_step : int
            Step between points to plot
        out_path : str, optional
            File path to save plot to, can be a .html or static image,
            by default None
        kwargs : dict
            Additional kwargs for plotly.express.imshow
        """
        fig = px.imshow(self.mask[::plot_step, ::plot_step],
                        color_continuous_scale=cmap, **kwargs)
        fig.update_layout(font=dict(family="Arial", size=18, color="black"))

        if out_path is not None:
            SummaryPlots._save_plotly(fig, out_path)

        fig.show()

    @classmethod
    def plot(cls, mask, out_dir, plot_type='plotly', cmap='Viridis',
             plot_step=100, **kwargs):
        """
        Plot exclusions mask and save to out_dir

        Parameters
        ----------
        mask : ndarray
            ndarray of final exclusions mask
        out_dir : str
            Output directory to save plots to
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        plot_step : int
            Step between points to plot
        kwargs : dict
            Additional plotting kwargs
        """
        excl_mask = cls(mask)
        if plot_type == 'plot':
            out_path = 'exclusions_mask.png'
            out_path = os.path.join(out_dir, out_path)
            excl_mask.exclusions_plot(cmap=cmap.lower(),
                                      plot_step=plot_step,
                                      out_path=out_path,
                                      **kwargs)
        elif plot_type == 'plotly':
            out_path = 'exclusions_mask.html'
            out_path = os.path.join(out_dir, out_path)
            excl_mask.exclusions_plotly(cmap=cmap.capitalize(),
                                        plot_step=plot_step,
                                        out_path=out_path,
                                        **kwargs)
        else:
            msg = ("plot_type must be 'plot' or 'plotly' but {} was given"
                   .format(plot_type))
            logger.error(msg)
            raise ValueError(msg)
